#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import itertools

import numpy as np
import pytest

torch = pytest.importorskip("torch")

import torch.nn as nn
import torch.nn.functional as F

from coremltools.converters.mil.mil import types
from coremltools.converters.mil.mil import Builder as mb
from coremltools.converters.mil.mil import Function, get_new_symbol
from coremltools.converters.mil.mil.var import Var
from coremltools.converters.mil.testing_reqs import random_gen

from .. import ops
from ..converter import TorchConverter, TranscriptionContext
from ..internal_graph import InternalTorchIRNode


class TestTorchOps:
    """Class containing tests for converting TorchIR -> CoreML ops.

    These tests interface with only the InternalTorchIRGraph and do not
    build a torch module. Thus, they are much faster then the numerical tests.
    However, for some ops it is necessary to use the torch module to verify
    numerical output so they are placed the numerical tests.

    NOTE: Confused where @context is coming from? Its from the pytest fixture defined below.
    """

    @pytest.fixture
    def context(self):
        return TranscriptionContext()

    @pytest.fixture
    def set_random_seeds(self):
        torch.manual_seed(1)
        np.random.seed(1)

    @pytest.mark.parametrize("dtype", [torch.bool, torch.float, torch.int])
    def test_constant(self, context, dtype):
        test_data = torch.ones(1, dtype=dtype)
        node = InternalTorchIRNode(
            attr={"value": test_data}, kind="constant", inputs=[], outputs=["1"]
        )
        ssa = self._construct_test_graph(context, ops.constant, node, "1")
        assert np.allclose(test_data, ssa.val)
        assert test_data.shape == ssa.shape

    def test_constant_magic(self, context):
        test_val = ops.PYTORCH_MAGIC_DEFAULT
        node = InternalTorchIRNode(
            attr={"value": test_val}, kind="constant", inputs=[], outputs=["1"]
        )
        ssa = self._construct_test_graph(context, ops.constant, node, "1")
        # We expect the magic default to get converted to None
        assert ssa is None

    @staticmethod
    def _gen_constants(size, vals):
        """Helper function. Generates a list of internal constant nodes.

        Arguments:
            size: number of constants to generate
            vals: Either a list of values for each constant or one value used for all constants."""
        is_list = isinstance(vals, list)
        if is_list:
            if len(vals) != size:
                raise ValueError("len(@vals): {} != size: {}".format(len(vals), size))
        constants = []
        for index in range(size):
            if is_list:
                val = vals[index]
            else:
                val = vals
            constants.append(
                InternalTorchIRNode(
                    attr={"value": val},
                    kind="constant",
                    inputs=[],
                    outputs=[str(index)],
                )
            )
        input_list = [str(i) for i in range(size)]
        output_name = str(len(input_list))
        return constants, input_list, output_name

    @staticmethod
    def _construct_test_graph(
        context, test_op, test_node, output_name=None, graph_inputs=None, constants=None
    ):
        """ Construct an Function for the given @graph_inputs, @constants,
            and @test_node. Returns the output of the graph, which is the ssa
            Var of the given @output_name.
        """
        if graph_inputs is None:
            graph_inputs = {}
        if constants is None:
            constants = []

        with Function(inputs=graph_inputs) as ssa_func:
            for name in ssa_func.inputs.keys():
                context.add(ssa_func.inputs[name])
            for node in constants:
                ops.constant(context, node)
            test_op(context, test_node)

        ssa = None
        if output_name:
            ssa = context[output_name]
        return ssa

    def _test_elementwise_binary(
        self, context, op_name, op, test_input, num_constants, expected_result
    ):
        """Helper function, runs op on test input and compares against expected result"""
        constants, input_list, output_name = self._gen_constants(
            num_constants, test_input
        )
        eb_node = InternalTorchIRNode(
            kind=op_name, inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, op, eb_node, output_name, constants=constants
        )
        np.testing.assert_allclose(expected_result, ssa.val, atol=1e-6)

    def _test_cast(self, context, test_val, op_kind, op_func, python_type):
        constants, input_list, output_name = self._gen_constants(1, [test_val])
        node = InternalTorchIRNode(
            kind=op_kind, inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, op_func, node, output_name, constants=constants
        )
        assert ssa.val == python_type(test_val)

    def test_add(self, context):
        test_input_1 = np.random.rand(2, 3)
        test_input_2 = np.random.rand(2, 3)
        scale_factor = 1
        self._test_elementwise_binary(
            context,
            "Add",
            ops.add,
            [test_input_1, test_input_2, scale_factor],
            3,
            test_input_1 + test_input_2,
        )

    def test_add_no_scale_factor(self, context):
        test_input_1 = np.random.rand(2, 3)
        test_input_2 = np.random.rand(2, 3)
        self._test_elementwise_binary(
            context,
            "Add",
            ops.add,
            [test_input_1, test_input_2],
            2,
            test_input_1 + test_input_2,
        )

    @pytest.mark.parametrize(
        "test_input_1, test_input_2",
        [(np.random.rand(3, 2), np.random.rand(3, 2)), (np.random.rand(3, 2), 5),],
    )
    def test_sub(self, context, test_input_1, test_input_2):
        scale_factor = 1
        self._test_elementwise_binary(
            context,
            "Sub",
            ops.sub,
            [test_input_1, test_input_2, scale_factor],
            3,
            test_input_1 - test_input_2,
        )

    @pytest.mark.parametrize(
        "test_input_1, test_input_2",
        [(np.random.rand(3, 2), np.random.rand(3, 2)), (np.random.rand(3, 2), 5),],
    )
    def test_rsub(self, context, test_input_1, test_input_2):
        scale_factor = 1
        self._test_elementwise_binary(
            context,
            "rsub",
            ops.sub,
            [test_input_1, test_input_2, scale_factor],
            3,
            # Note the reversal of arg ordering relative to 'sub'
            test_input_2 - test_input_1,
        )

    def test_mul(self, context):
        test_input_1 = np.random.rand(3, 2)
        test_input_2 = np.random.rand(3, 2)
        self._test_elementwise_binary(
            context,
            "Mul",
            ops.mul,
            [test_input_1, test_input_2],
            2,
            test_input_1 * test_input_2,
        )

    def test_div(self, context):
        test_input_1 = np.random.rand(3, 2)
        test_input_2 = np.random.rand(3, 2)
        self._test_elementwise_binary(
            context,
            "Div",
            ops.div,
            [test_input_1, test_input_2],
            2,
            np.divide(test_input_1, test_input_2),
        )

    def test_floor_divide(self, context):
        test_input_1 = np.random.randint(low=1, high=100, size=(3, 2))
        test_input_2 = np.random.randint(low=1, high=100, size=(3, 2))
        self._test_elementwise_binary(
            context,
            "floor_divide",
            ops.floor_divide,
            [test_input_1, test_input_2],
            2,
            np.floor_divide(test_input_1, test_input_2),
        )

    def test_pow(self, context):
        test_input_1 = np.random.rand(3, 2)
        test_input_2 = np.random.rand(3, 2)
        self._test_elementwise_binary(
            context,
            "Pow",
            ops.pow_,
            [test_input_1, test_input_2],
            2,
            np.power(test_input_1, test_input_2),
        )

    def test_eq(self, context):
        test_input_1 = torch.zeros([2, 3, 4, 5, 6]).float()
        test_input_2 = torch.ones([2, 3, 4, 5, 6]).float()
        test_input_2[0][0][0][0][0] = 0
        expected_output = (test_input_1 == test_input_2).float()

        self._test_elementwise_binary(
            context, "Eq", ops.eq, [test_input_1, test_input_2], 2, expected_output
        )

    def test_ne(self, context):
        test_input_1 = torch.zeros([2, 3, 4, 5, 6]).float()
        test_input_2 = torch.ones([2, 3, 4, 5, 6]).float()
        test_input_2[0][0][0][0][0] = 0
        expected_output = (test_input_1 != test_input_2).float()

        self._test_elementwise_binary(
            context, "ne", ops.ne, [test_input_1, test_input_2], 2, expected_output
        )

    def test_le(self, context):
        test_input_1 = torch.zeros([2, 3, 4, 5, 6]).float()
        test_input_2 = torch.ones([2, 3, 4, 5, 6]).float()
        test_input_2[0][0][0][0][0] = 0
        expected_output = (test_input_1 <= test_input_2).float()

        self._test_elementwise_binary(
            context, "Le", ops.le, [test_input_1, test_input_2], 2, expected_output
        )

    def test_lt(self, context):
        test_input_1 = torch.zeros([2, 3, 4, 5, 6]).float()
        test_input_2 = torch.ones([2, 3, 4, 5, 6]).float()
        test_input_2[0][0][0][0][0] = 0
        expected_output = (test_input_1 < test_input_2).float()

        self._test_elementwise_binary(
            context, "Lt", ops.lt, [test_input_1, test_input_2], 2, expected_output
        )

    def test_ge(self, context):
        test_input_1 = torch.zeros([2, 3, 4, 5, 6]).float()
        test_input_2 = torch.ones([2, 3, 4, 5, 6]).float()
        test_input_2[0][0][0][0][0] = 0
        expected_output = (test_input_1 >= test_input_2).float()

        self._test_elementwise_binary(
            context, "Ge", ops.ge, [test_input_1, test_input_2], 2, expected_output
        )

    def test_gt(self, context):
        test_input_1 = torch.zeros([2, 3, 4, 5, 6]).float()
        test_input_2 = torch.ones([2, 3, 4, 5, 6]).float()
        test_input_2[0][0][0][0][0] = 0
        expected_output = (test_input_1 > test_input_2).float()

        self._test_elementwise_binary(
            context, "Gt", ops.gt, [test_input_1, test_input_2], 2, expected_output
        )

    @pytest.mark.parametrize(
        "size, array_type",
        itertools.product(
            [1, 5, 7],
            [
                ("ListConstruct", ops.listconstruct),
                ("TupleConstruct", ops.tupleconstruct),
            ],
        ),
    )
    def test_arrayconstruct_scalars(self, context, size, array_type):
        constant_vals = list(range(size))
        array_kind = array_type[0]
        array_op = array_type[1]
        constants, input_list, output_name = self._gen_constants(size, constant_vals)
        ac_node = InternalTorchIRNode(
            kind=array_kind, inputs=input_list, outputs=[output_name],
        )
        ssa = self._construct_test_graph(
            context, array_op, ac_node, output_name, constants=constants
        )
        expected_val = np.arange(size)
        np.testing.assert_equal(ssa.shape, (size,))
        np.testing.assert_array_equal(ssa.val, expected_val)

    @pytest.mark.parametrize(
        "shape1, shape2, array_type",
        itertools.product(
            [(1, 2), (3, 4, 5), (2,)],
            [(2, 1), (1, 4, 5), (3,)],
            [
                ("ListConstruct", ops.listconstruct),
                ("TupleConstruct", ops.tupleconstruct),
            ],
        ),
    )
    def test_arrayconstruct_nonscalar(self, context, shape1, shape2, array_type):
        tensor1 = torch.rand(shape1)
        tensor2 = torch.rand(shape2)
        array_kind = array_type[0]
        array_op = array_type[1]
        constants, input_list, output_name = self._gen_constants(2, [tensor1, tensor2])
        ac_node = InternalTorchIRNode(
            kind=array_kind, inputs=input_list, outputs=[output_name],
        )
        ssa = self._construct_test_graph(
            context, array_op, ac_node, output_name, constants=constants
        )
        expected_val = (tensor1.numpy(), tensor2.numpy())
        np.testing.assert_equal(len(ssa), 2)
        for x, y in zip(ssa, expected_val):
            np.testing.assert_allclose(x.val, y)

    @pytest.mark.parametrize(
        "input_shape, dim0, dim1",
        [
            x
            for x in itertools.product(
                [(1, 2, 3), (1, 2, 3, 4), (1, 2, 3, 4, 5)], [0, 1, -1], [0, 2, -2],
            )
        ]
        + [((1, 2), None, None)],
    )
    def test_transpose(self, context, input_shape, dim0, dim1):
        test_input = torch.rand(input_shape)

        constant_list = [test_input]
        if len(input_shape) > 2:
            constant_list += [dim0, dim1]
            kind = "transpose"
            expected_result = torch.transpose(test_input, dim0, dim1)
        else:
            kind = "t"
            expected_result = test_input.t()

        constants, input_list, output_name = self._gen_constants(
            len(constant_list), constant_list
        )
        transpose_node = InternalTorchIRNode(
            kind=kind, inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.transpose, transpose_node, output_name, constants=constants,
        )
        np.testing.assert_array_equal(expected_result.shape, ssa.shape)
        np.testing.assert_allclose(expected_result, ssa.val)

    @pytest.mark.parametrize(
        "dim1, dim2, dim3", itertools.product([1, 2, 5], [2, 5, 10], [1, 2, 5]),
    )
    def test_matmul(self, context, dim1, dim2, dim3):
        mat1 = torch.rand((dim1, dim2))
        mat2 = torch.rand((dim2, dim3))
        constant_vals = [
            mat1,
            mat2,
        ]
        constants, input_list, output_name = self._gen_constants(2, constant_vals)

        matmul_node = InternalTorchIRNode(
            kind="matmul", inputs=input_list, outputs=[output_name],
        )

        ssa = self._construct_test_graph(
            context, ops.matmul, matmul_node, output_name, constants=constants
        )
        expected_result = torch.matmul(mat1, mat2).detach().numpy()
        assert np.allclose(expected_result, ssa.val)

    @pytest.mark.parametrize(
        "input_shape, axis, expected_shape",
        [
            ((1, 2), None, (2,)),
            ((1, 2), 0, (2,)),
            ((1, 2, 1), None, (2,)),
            ((1, 2, 1, 1), None, (2,)),
            ((1, 2, 1, 1), 2, (1, 2, 1)),
            ((1, 2, 1, 1, 1), None, (2,)),
        ],
    )
    def test_squeeze(self, context, input_shape, axis, expected_shape):
        test_data = torch.rand(input_shape)
        if axis is None:
            constants, input_list, output_name = self._gen_constants(1, test_data)
        else:
            constants, input_list, output_name = self._gen_constants(
                2, [test_data, axis]
            )
        squeeze_node = InternalTorchIRNode(
            kind="Squeeze", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.squeeze, squeeze_node, output_name, constants=constants
        )
        if axis is None:
            expected_result = torch.squeeze(test_data)
        else:
            expected_result = torch.squeeze(test_data, axis)
        assert np.allclose(expected_result, ssa.val)
        assert expected_result.size() == torch.Size(expected_shape)

    @pytest.mark.parametrize(
        "input_shape, axis, expected_shape",
        [
            ((2,), 0, (1, 2)),
            ((2,), 1, (2, 1)),
            ((2,), -1, (2, 1)),
            ((2, 3), 1, (2, 1, 3)),
        ],
    )
    def test_unsqueeze(self, context, input_shape, axis, expected_shape):
        test_data = torch.rand(input_shape)
        constants, input_list, output_name = self._gen_constants(2, [test_data, axis])
        unsqueeze_node = InternalTorchIRNode(
            kind="Unsqueeze", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.unsqueeze, unsqueeze_node, output_name, constants=constants
        )
        expected_result = torch.unsqueeze(test_data, axis)
        assert np.allclose(expected_result, ssa.val)
        assert expected_result.size() == torch.Size(expected_shape)

    @pytest.mark.parametrize(
        "input_shape, start, end",
        [
            ((2, 1, 1, 2), 1, 3),
            ((2, 2, 1, 1), 1, -2),
            ((1, 1, 1), 0, 2),
            ((1, 2), 0, 1),
            ((1, 2), 1, 1),
            ((1, 1), 1, -1),
            ((1,), 0, 0),
        ],
    )
    def test_flatten(self, context, input_shape, start, end):
        test_data = torch.rand(input_shape)
        constants, input_list, output_name = self._gen_constants(
            3, [test_data, start, end]
        )
        flatten_node = InternalTorchIRNode(
            kind="Flatten", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.flatten, flatten_node, output_name, constants=constants
        )
        expected_result = torch.flatten(test_data, start, end)
        assert np.allclose(expected_result, ssa.val)

    @pytest.mark.parametrize(
        "start, end", [(0, -5), (100, 2), (2, 100), (-3, -4),],
    )
    def test_flatten_exception(self, context, start, end):
        test_data = torch.rand(1, 1, 1, 1)
        constants, input_list, output_name = self._gen_constants(
            3, [test_data, start, end]
        )
        flatten_node = InternalTorchIRNode(
            kind="Flatten", inputs=input_list, outputs=[output_name]
        )
        with pytest.raises(ValueError):
            self._construct_test_graph(
                context, ops.flatten, flatten_node, output_name, constants=constants,
            )

    @pytest.mark.parametrize(
        "input_shape", [(2, 3), (2, 3, 4), (2, 3, 4, 5), (2, 3, 4, 5, 6),],
    )
    def test_permute(self, context, input_shape):
        test_data = torch.rand(*input_shape)
        permutation = list(range(len(input_shape)))
        np.random.shuffle(permutation)
        constants, input_list, output_name = self._gen_constants(
            2, [test_data, permutation]
        )
        permute_node = InternalTorchIRNode(
            kind="Permute", inputs=input_list, outputs=[output_name],
        )
        ssa = self._construct_test_graph(
            context, ops.permute, permute_node, output_name, constants=constants
        )
        expected_result = test_data.permute(*permutation)
        assert expected_result.shape == ssa.shape

    @pytest.mark.parametrize(
        "in_features, out_features, scaling",
        itertools.product([10, 25, 100], [3, 6], [1.0, 0.5]),
    )
    def test_addmm(self, context, in_features, out_features, scaling):
        input_data = torch.rand((1, in_features))
        weight_data = torch.rand((in_features, out_features))
        bias_data = torch.rand((out_features))
        constant_vals = [
            scaling,
            input_data,
            weight_data,
            bias_data,
        ]
        constants, _, output_name = self._gen_constants(4, constant_vals)

        addmm_node = InternalTorchIRNode(
            kind="addmm", inputs=["3", "1", "2", "0", "0"], outputs=[output_name],
        )

        ssa = self._construct_test_graph(
            context, ops.addmm, addmm_node, output_name, constants=constants
        )
        torch_linear = nn.Linear(in_features=in_features, out_features=out_features,)
        expected_shape = tuple(torch_linear(input_data).shape)
        assert expected_shape == ssa.shape

    @pytest.mark.parametrize(
        "height, width, kernel_size, stride, padding, dilation",
        itertools.product([5, 6], [5, 7], [1, 3], [1, 3], [1, 3], [1, 3]),
    )
    def test_convolution2d(
        self,
        context,
        height,
        width,
        kernel_size,
        stride,
        padding,
        dilation,
        groups=1,
        in_channels=1,
        out_channels=2,
    ):
        test_input = torch.rand(1, in_channels, height, width)
        constant_vals = [
            1,  # None argument
            test_input,
            np.random.rand(
                out_channels, in_channels, kernel_size, kernel_size
            ),  # weights
            np.random.rand(out_channels),  # bias
            np.array([stride, stride]),
            np.array([padding, padding]),
            np.array([dilation, dilation]),
            False,  # transposed
            np.array([0, 0]),  # output_pad
            groups,
        ]
        constants, _, output_name = self._gen_constants(
            len(constant_vals), constant_vals
        )
        # For reference, the values for `kind` and `inputs` indices are determined from the definition for Torch's
        # `at::_convolution` used for all convolutions. The link below is approximately correct at the time of writing.
        # https://github.com/pytorch/pytorch/blob/bd604mb5b7ae4f6388aca461891d620b0d485fbb/aten/src/ATen/native/Convolution.cpp#L544
        conv_node = InternalTorchIRNode(
            kind="_convolution",
            inputs=["1", "2", "3", "4", "5", "6", "7", "8", "9", "0", "0", "0"],
            outputs=[output_name],
        )

        ssa = self._construct_test_graph(
            context, ops._convolution, conv_node, output_name, constants=constants
        )
        torch_conv = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
        )
        expected_shape = tuple(torch_conv(test_input).shape)
        assert ssa.val == None
        assert expected_shape == ssa.shape

    @pytest.mark.parametrize(
        "depth, height, width, kernel_size, stride, padding, dilation, groups",
        itertools.product(
            [5, 5],
            [5, 6],
            [5, 7],
            [1, 3],
            [(1, 1, 1), (3, 2, 1)],
            [(1, 1, 1), (1, 3, 2)],
            [(1, 1, 1), (1, 2, 3)],
            [
                1,
                -1,
            ],  # -1 groups indicates it should be set to the number of input channels for depthwise convolution
        ),
    )
    def test_convolution3d(
        self,
        context,
        depth,
        height,
        width,
        kernel_size,
        stride,
        padding,
        dilation,
        groups,
        in_channels=2,
        out_channels=4,
    ):
        if groups == -1:
            groups = in_channels
        test_input = torch.rand(1, in_channels, depth, height, width)
        constant_vals = [
            1,  # None argument
            test_input,
            np.random.rand(
                out_channels,
                in_channels // groups,
                kernel_size,
                kernel_size,
                kernel_size,
            ),  # weights
            np.random.rand(out_channels),  # bias
            # PyTorch's Conv3d accepts either an int (for all dimensions) or a 3-tuple of ints (one per dimension)
            np.array([stride[0], stride[1], stride[2]]),
            np.array([padding[0], padding[1], padding[2]]),
            np.array([dilation[0], dilation[1], dilation[2]]),
            False,  # transposed
            np.array([0, 0, 0]),  # out_pad
            groups,
        ]
        constants, _, output_name = self._gen_constants(
            len(constant_vals), constant_vals
        )
        # For reference, the values for `kind` and `inputs` indices are determined from the definition for Torch's
        # `at::_convolution` used for all convolutions. The link below is approximately correct at the time of writing.
        # https://github.com/pytorch/pytorch/blob/bd604mb5b7ae4f6388aca461891d620b0d485fbb/aten/src/ATen/native/Convolution.cpp#L544
        conv_node = InternalTorchIRNode(
            kind="_convolution",
            inputs=["1", "2", "3", "4", "5", "6", "7", "8", "9", "0", "0", "0"],
            outputs=[output_name],
        )

        ssa = self._construct_test_graph(
            context, ops._convolution, conv_node, output_name, constants=constants
        )
        torch_conv = nn.Conv3d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
        )
        expected_result = torch_conv(test_input)
        expected_shape = tuple(expected_result.shape)
        assert ssa.val is None
        assert expected_shape == ssa.shape

    @pytest.mark.parametrize(
        "height, width, kernel_size, stride, padding, dilation",
        itertools.product([5, 6], [5, 7], [1, 3], [2, 3], [0, 1], [1, 3]),
    )
    def test_convolution_transpose2d(
        self,
        context,
        height,
        width,
        kernel_size,
        stride,
        padding,
        dilation,
        groups=1,
        in_channels=1,
        out_channels=2,
    ):
        test_input = torch.rand(1, in_channels, height, width)

        constant_vals = [
            np.random.rand(
                in_channels, out_channels, kernel_size, kernel_size
            ),  # weights
            np.random.rand(out_channels),  # bias
            np.array([stride, stride]),
            np.array([padding, padding]),
            np.array([dilation, dilation]),
            True,  # transposed,
            np.array([0, 0]),  # output_pad
            groups,
            False,
            False,
            False,
        ]
        graph_inputs = {"input": mb.placeholder(test_input.shape, dtype=types.float)}

        constants, input_list, output_name = self._gen_constants(
            len(constant_vals), constant_vals
        )
        conv_node = InternalTorchIRNode(
            kind="_convolution", inputs=["input"] + input_list, outputs=[output_name],
        )

        ssa = self._construct_test_graph(
            context,
            ops._convolution,
            conv_node,
            output_name,
            constants=constants,
            graph_inputs=graph_inputs,
        )
        torch_conv = nn.ConvTranspose2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
        )
        expected_shape = tuple(torch_conv(test_input).shape)
        assert ssa.val == None
        assert expected_shape == ssa.shape

    @pytest.mark.parametrize(
        "input_shape, dim, keepdim",
        itertools.product([(3, 20, 20), (1, 50, 50)], [0, 1, 2, [0, 2]], [True, False]),
    )
    def test_mean(self, context, input_shape, dim, keepdim):
        test_input = torch.rand(*input_shape)

        constants, input_list, output_name = self._gen_constants(
            4, [test_input, dim, keepdim, None]
        )
        mean_node = InternalTorchIRNode(
            kind="mean", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.mean, mean_node, output_name, constants=constants
        )
        expected_result = torch.mean(test_input, dim, keepdim)
        assert np.allclose(expected_result, ssa.val)

    def test_mean_no_dims(self, context):
        test_input = torch.rand((3, 20, 20))

        constants, input_list, output_name = self._gen_constants(2, [test_input, None])
        mean_node = InternalTorchIRNode(
            kind="mean", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.mean, mean_node, output_name, constants=constants
        )
        expected_result = torch.mean(test_input)
        assert np.allclose(expected_result, ssa.val)

    def test_embedding(self, context):
        EMBEDDING_DIMENSION = 10
        NUM_EMBEDDINGS = 20
        input_shape = (NUM_EMBEDDINGS, EMBEDDING_DIMENSION)
        # size is arbitrary for indices
        indices = np.random.randint(NUM_EMBEDDINGS, size=100)
        test_input = torch.rand(input_shape)
        constants, input_list, output_name = self._gen_constants(
            2, [test_input, indices]
        )
        gather_node = InternalTorchIRNode(
            kind="embedding", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.embedding, gather_node, output_name, constants=constants
        )
        torch_embedding = nn.Embedding.from_pretrained(test_input)
        expected_result = torch_embedding(torch.LongTensor(indices))
        assert np.allclose(expected_result, ssa.val)

    @pytest.mark.parametrize(
        "dim", [0, 1, 2, 3, 4],
    )
    def test_size(self, context, dim):
        test_input = torch.rand(1, 2, 3, 4, 5)

        graph_inputs = {"input": mb.placeholder(test_input.shape, dtype=types.float)}
        constants, input_list, output_name = self._gen_constants(1, [dim])
        size_node = InternalTorchIRNode(
            kind="size", inputs=["input"] + input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context,
            ops.size,
            size_node,
            output_name,
            constants=constants,
            graph_inputs=graph_inputs,
        )
        expected_result = test_input.shape[dim]
        assert expected_result == ssa.val

    @pytest.mark.parametrize(
        "dim", [0, 1],
    )
    def test_size_symbolic(self, context, dim):
        test_shape = (3, get_new_symbol())
        graph_inputs = {"input": mb.placeholder(shape=test_shape, dtype=types.float)}
        constants, input_list, output_name = self._gen_constants(1, [dim])
        size_node = InternalTorchIRNode(
            kind="size", inputs=["input"] + input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context,
            ops.size,
            size_node,
            output_name,
            constants=constants,
            graph_inputs=graph_inputs,
        )
        expected_result = test_shape[dim]
        assert expected_result == ssa.sym_val

    @pytest.mark.parametrize(
        "input_size, shape",
        itertools.product([(5, 12), (1, 4, 15), (3, 5, 4)], [(3, 20), (-1, 6), (60,)],),
    )
    def test_view(self, context, input_size, shape):
        test_input = torch.rand(input_size)

        constants, input_list, output_name = self._gen_constants(2, [test_input, shape])
        view_node = InternalTorchIRNode(
            kind="view", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.view, view_node, output_name, constants=constants
        )
        expected_result = test_input.view(shape)
        assert np.allclose(expected_result, ssa.val)

    @pytest.mark.parametrize(
        "input_shape, output_shape",
        itertools.product(
            [(1, 3, 15, 15), (1, 1, 2, 2), (1, 3, 10, 10)], [(1, 1), (2, 2), (2, 1)],
        ),
    )
    def test_adaptive_avg_pool2d(self, context, input_shape, output_shape):
        test_input = torch.rand(input_shape)

        constants, input_list, output_name = self._gen_constants(
            2, [test_input, output_shape]
        )

        adaptive_avg_pool2d_node = InternalTorchIRNode(
            kind="adaptive_avg_pool2d", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context,
            ops.adaptive_avg_pool2d,
            adaptive_avg_pool2d_node,
            output_name,
            constants=constants,
        )
        expected_result = torch._adaptive_avg_pool2d(test_input, output_shape)
        expected_shape = tuple(expected_result.shape)
        assert expected_shape == ssa.shape
        # We only expect numerical output when reducing to global average.
        if output_shape == (1, 1):
            assert np.allclose(expected_result, ssa.val)

    def test_adaptive_avg_pool2d_exception(self, context):
        # For this test, the input tensor HW channels are dynamic.
        input_shape = [1, 3, get_new_symbol(), get_new_symbol()]
        graph_inputs = {"input": mb.placeholder(input_shape, dtype=types.float)}
        constants, input_list, output_name = self._gen_constants(1, [(2, 1)])
        adaptive_avg_pool2d_node = InternalTorchIRNode(
            kind="adaptive_avg_pool2d",
            inputs=["input"] + input_list,
            outputs=[output_name],
        )
        with pytest.raises(ValueError):
            ssa = self._construct_test_graph(
                context,
                ops.adaptive_avg_pool2d,
                adaptive_avg_pool2d_node,
                output_name,
                constants=constants,
                graph_inputs=graph_inputs,
            )

    @pytest.mark.parametrize("input_shape", [(1, 3, 15, 15), (1, 1, 1, 1)])
    def test_batch_norm(self, context, input_shape):
        test_input = torch.rand(input_shape)
        channels = input_shape[1]
        constants, input_list, output_name = self._gen_constants(
            9,
            [
                torch.rand(input_shape),  # input
                torch.rand(channels),  # weight
                torch.rand(channels),  # bias
                torch.rand(channels),  # running mean
                torch.rand(channels),  # running var
                0,  # training
                0.1,  # momentum
                1e-6,  # eps
                1,  # cudnn_enabled
            ],
        )

        batch_norm_node = InternalTorchIRNode(
            kind="batch_norm", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.batch_norm, batch_norm_node, output_name, constants=constants
        )
        assert ssa.val == None
        assert ssa.shape == tuple(test_input.shape)

    @pytest.mark.parametrize("input_shape", [(1, 3, 15, 15), (1, 1, 1, 1)])
    def test_instance_norm(self, context, input_shape):
        test_input = torch.rand(input_shape)
        channels = input_shape[1]
        constants, input_list, output_name = self._gen_constants(
            9,
            [
                torch.rand(input_shape),  # input
                torch.rand(channels),  # weight
                torch.rand(channels),  # bias
                torch.rand(channels),  # running mean
                torch.rand(channels),  # running var
                0,  # training
                0.1,  # momentum
                1e-6,  # eps
                1,  # cudnn_enabled
            ],
        )

        instant_norm_node = InternalTorchIRNode(
            kind="instance_norm", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.instance_norm, instant_norm_node, output_name, constants=constants
        )
        assert ssa.val == None
        assert ssa.shape == tuple(test_input.shape)

    @pytest.mark.parametrize("axis", [1, 2, 3])
    def test_cat(self, context, axis):
        input_shape = (1, 3, 240, 320)

        test_input1 = torch.rand(input_shape)
        test_input2 = torch.rand(input_shape)
        const_input = torch.rand(input_shape)

        graph_inputs = {
            "input1": mb.placeholder(input_shape, dtype=types.float),
            "input2": mb.placeholder(input_shape, dtype=types.float),
        }
        dim_node = InternalTorchIRNode(
            attr={"value": axis}, kind="constant", inputs=[], outputs=["0"],
        )
        const_tensor_node = InternalTorchIRNode(
            attr={"value": const_input.numpy()},
            kind="constant",
            inputs=[],
            outputs=["1"],
        )
        listconstruct_node = InternalTorchIRNode(
            kind="listconstruct", inputs=["1", "input1", "input2"], outputs=["2"]
        )
        cat_node = InternalTorchIRNode(
            kind="cat", inputs=["2", "0"], outputs=["output"]
        )

        with Function(inputs=graph_inputs) as ssa_func:
            context.add(ssa_func.inputs["input1"])
            context.add(ssa_func.inputs["input2"])
            ops.constant(context, dim_node)
            ops.constant(context, const_tensor_node)
            ops.listconstruct(context, listconstruct_node)
            ops.cat(context, cat_node)

        ssa = context["output"]
        expected_result = torch.cat(
            (const_input, test_input1, test_input2), dim=axis
        ).numpy()
        assert np.allclose(expected_result.shape, ssa.shape)

    @pytest.mark.parametrize("axis", [0, 1, 2, 3, 4])
    def test_stack(self, context, axis):
        input_shape = (1, 3, 240, 320)

        test_input1 = torch.rand(input_shape)
        test_input2 = torch.rand(input_shape)
        const_input = torch.rand(input_shape)

        graph_inputs = {
            "input1": mb.placeholder(input_shape, dtype=types.float),
            "input2": mb.placeholder(input_shape, dtype=types.float),
        }
        dim_node = InternalTorchIRNode(
            attr={"value": axis}, kind="constant", inputs=[], outputs=["0"],
        )
        const_tensor_node = InternalTorchIRNode(
            attr={"value": const_input.numpy()},
            kind="constant",
            inputs=[],
            outputs=["1"],
        )
        listconstruct_node = InternalTorchIRNode(
            kind="listconstruct", inputs=["1", "input1", "input2"], outputs=["2"]
        )
        stack_node = InternalTorchIRNode(
            kind="stack", inputs=["2", "0"], outputs=["output"]
        )

        with Function(inputs=graph_inputs) as ssa_func:
            context.add(ssa_func.inputs["input1"])
            context.add(ssa_func.inputs["input2"])
            ops.constant(context, dim_node)
            ops.constant(context, const_tensor_node)
            ops.listconstruct(context, listconstruct_node)
            ops.stack(context, stack_node)

        ssa = context["output"]
        expected_result = np.stack((const_input, test_input1, test_input2), axis=axis)
        assert np.allclose(expected_result.shape, ssa.shape)

    def test_item(self, context):
        const_val = 0
        constants, input_list, output_name = self._gen_constants(1, [const_val])
        item_node = InternalTorchIRNode(
            kind="item", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.item, item_node, output_name, constants=constants
        )
        assert ssa.val == const_val

    def test_item_exception(self, context):
        const_val = [0, 1]
        constants, input_list, output_name = self._gen_constants(1, [const_val])
        item_node = InternalTorchIRNode(
            kind="item", inputs=input_list, outputs=[output_name]
        )
        with pytest.raises(ValueError):
            ssa = self._construct_test_graph(
                context, ops.item, item_node, output_name, constants=constants,
            )

    @pytest.mark.parametrize("test_val", [1, 1.5, False])
    def test_bool(self, context, test_val):
        self._test_cast(context, test_val, "bool", ops._bool, bool)

    @pytest.mark.parametrize("test_val", [1, 1.5, -0.3])
    def test_int(self, context, test_val):
        self._test_cast(context, test_val, "int", ops._int, int)

    @pytest.mark.parametrize("input_shape", [(1, 3, 15, 15), (1, 1, 1, 1)])
    def test_layer_norm(self, context, input_shape):
        graph_inputs = {"input": mb.placeholder(input_shape, dtype=types.float)}
        constants, input_list, output_name = self._gen_constants(
            5,
            [
                input_shape,  # normalized shape
                torch.rand(*input_shape),  # weight
                torch.rand(*input_shape),  # running bias
                1e-6,
                1,  # cudnn enabled
            ],
        )

        layer_norm_node = InternalTorchIRNode(
            kind="layer_norm", inputs=["input"] + input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context,
            ops.layer_norm,
            layer_norm_node,
            output_name,
            graph_inputs=graph_inputs,
            constants=constants,
        )
        assert ssa.val == None
        assert ssa.shape == input_shape

    @pytest.mark.parametrize("shape", [(1, 2), (2, 3, 4, 5), (3, 4, 5),])
    def test_ones(self, context, shape):
        constants, constant_input_list, output_name = self._gen_constants(
            6, [shape, 1, 1, 1, 1, 1]
        )
        ones_node = InternalTorchIRNode(
            kind="ones", inputs=constant_input_list, outputs=[output_name],
        )
        ssa = self._construct_test_graph(
            context, ops.ones, ones_node, output_name, constants=constants,
        )
        assert ssa.shape == shape

    @pytest.mark.parametrize("input_shape", [(1, 2), (2, 3, 4, 5), (3, 4, 5),])
    def test_ones_like(self, context, input_shape):
        graph_inputs = {"input": mb.placeholder(input_shape, dtype=types.float)}
        constants, constant_input_list, output_name = self._gen_constants(5, 1)
        ones_node = InternalTorchIRNode(
            kind="ones_like",
            inputs=["input"] + constant_input_list,
            outputs=[output_name],
        )
        ssa = self._construct_test_graph(
            context,
            ops.ones_like,
            ones_node,
            output_name,
            graph_inputs=graph_inputs,
            constants=constants,
        )
        assert ssa.shape == input_shape

    @pytest.mark.parametrize(
        "input_size, dim, index",
        itertools.product(
            [(13, 43, 10), (39, 14, 11, 9)], [0, 1, 2], [0, 1, 3, 8, -1],
        ),
    )
    def test_select(self, context, input_size, dim, index):
        graph_inputs = {"input1": mb.placeholder(input_size, dtype=types.float)}
        constants, constant_input_list, output_name = self._gen_constants(
            2, [dim, index]
        )
        select_node = InternalTorchIRNode(
            kind="select",
            inputs=["input1"] + constant_input_list,
            outputs=[output_name],
        )
        ssa = self._construct_test_graph(
            context,
            ops.select,
            select_node,
            output_name,
            graph_inputs=graph_inputs,
            constants=constants,
        )
        select_index = index
        if index < 0:
            select_index += input_size[dim]
        expected_shape = tuple(
            torch.rand(input_size)
            .index_select(dim, torch.tensor([select_index]))
            .squeeze(dim)
            .shape
        )
        assert np.allclose(ssa.shape, expected_shape)

    @pytest.mark.parametrize(
        "dynamic, test_tuple", itertools.product([True, False], [True, False])
    )
    def test_tuple_and_list_unpack(self, context, dynamic, test_tuple):
        """
            if @dynamic is True then packs up a dynamic input
            if @test_tuple is True tests tupleUnpack else tests listUnpack
        """
        if test_tuple:
            construct_op = ops.tupleconstruct
            construct_name = "TupleConstruct"
            unpack_name = "TupleUnpack"
        else:
            construct_op = ops.listconstruct
            construct_name = "ListConstruct"
            unpack_name = "ListUnpack"

        input_shape = (1, 2, 3)
        constant_vals = [str(i) for i in range(1, 6)]
        constants_unpacked = [str(i) for i in range(6, 11)]
        constants, input_list, _ = self._gen_constants(5, constant_vals)
        output_list = constants_unpacked[:]
        graph_inputs = {}
        if dynamic:
            graph_input_name = "input1"
            graph_inputs = {
                graph_input_name: mb.placeholder(input_shape, dtype=types.float)
            }
            input_list += [graph_input_name]
            output_list += [graph_input_name + "_out"]

        construct_node = InternalTorchIRNode(
            kind=construct_name, inputs=input_list, outputs=["construct"],
        )
        unpack_node = InternalTorchIRNode(
            kind=unpack_name, inputs=["construct"], outputs=output_list
        )
        with Function(inputs=graph_inputs) as ssa_func:
            if dynamic:
                context.add(ssa_func.inputs["input1"])
            for node in constants:
                ops.constant(context, node)
            construct_op(context, construct_node)
            ops.tupleunpack(context, unpack_node)

        ssa_constants = []
        for name in constants_unpacked:
            ssa_constants.append(context[name].val)
        assert ssa_constants == constant_vals

        if dynamic:
            ssa_dyanmic = context[graph_input_name + "_out"]
            assert ssa_dyanmic.val is None
            assert ssa_dyanmic.shape == input_shape

    def _test_pool(
        self, context, test_input, param_list, op_kind, op_func, expected_result
    ):
        constants, input_list, output_name = self._gen_constants(
            len(param_list) + 1, [test_input] + param_list,
        )

        pool_node = InternalTorchIRNode(
            kind=op_kind, inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, op_func, pool_node, output_name, constants=constants,
        )
        expected_shape = tuple(expected_result.shape)
        assert expected_shape == ssa.shape

    @pytest.mark.parametrize(
        "input_shape, kernel_size, stride, pad, include_pad, ceil_mode",
        itertools.product(
            [(1, 3, 15), (1, 1, 7), (1, 3, 10)],
            [1, 3],
            [1, 2],
            [0, 1],
            [True, False],
            [False, True],
        ),
    )
    def test_avg_pool1d(
        self, context, input_shape, kernel_size, stride, pad, include_pad, ceil_mode,
    ):
        if pad > kernel_size / 2:
            return
        test_input = torch.rand(input_shape)
        expected_result = F.avg_pool1d(
            test_input,
            kernel_size=kernel_size,
            stride=stride,
            padding=pad,
            ceil_mode=ceil_mode,
            count_include_pad=include_pad,
        )
        self._test_pool(
            context,
            test_input,
            [[kernel_size], [stride], [pad], ceil_mode, not include_pad],
            "avg_pool1d",
            ops.avg_pool1d,
            expected_result,
        )

    @pytest.mark.parametrize(
        "input_shape, kernel_size, stride, pad, include_pad, ceil_mode",
        itertools.product(
            [(1, 3, 15, 15), (1, 1, 7, 7), (1, 3, 10, 10)],
            [1, 3],
            [1, 2],
            [0, 1],
            [True, False],
            [False, True],
        ),
    )
    def test_avg_pool2d(
        self, context, input_shape, kernel_size, stride, pad, include_pad, ceil_mode,
    ):
        if pad > kernel_size / 2:
            return
        test_input = torch.rand(input_shape)
        expected_result = F.avg_pool2d(
            test_input,
            kernel_size=kernel_size,
            stride=stride,
            padding=pad,
            ceil_mode=ceil_mode,
            count_include_pad=include_pad,
        )
        self._test_pool(
            context,
            test_input,
            [
                [kernel_size, kernel_size],
                [stride, stride],
                [pad, pad],
                ceil_mode,
                not include_pad,
                None,
            ],
            "avg_pool2d",
            ops.avg_pool2d,
            expected_result,
        )

    @pytest.mark.parametrize(
        "input_shape, kernel_size, stride, pad, ceil_mode",
        itertools.product(
            [(1, 3, 15), (1, 1, 7), (1, 3, 10)], [1, 3], [1, 2], [0, 1], [False, True]
        ),
    )
    def test_max_pool1d(
        self, context, input_shape, kernel_size, stride, pad, ceil_mode
    ):
        if pad > kernel_size / 2:
            return
        test_input = torch.rand(input_shape)
        expected_result = F.max_pool1d(
            test_input,
            kernel_size=kernel_size,
            stride=stride,
            padding=pad,
            ceil_mode=ceil_mode,
        )
        self._test_pool(
            context,
            test_input,
            [[kernel_size], [stride], [pad], [1], ceil_mode],
            "max_pool1d",
            ops.max_pool1d,
            expected_result,
        )

    @pytest.mark.parametrize(
        "input_shape, kernel_size, stride, pad, ceil_mode",
        itertools.product(
            [(1, 3, 15, 15), (1, 1, 7, 7), (1, 3, 10, 10)],
            [1, 3],
            [1, 2],
            [0, 1],
            [False, True],
        ),
    )
    def test_max_pool2d(
        self, context, input_shape, kernel_size, stride, pad, ceil_mode,
    ):
        if pad > kernel_size / 2:
            return
        test_input = torch.rand(input_shape)
        expected_result = F.max_pool2d(
            test_input,
            kernel_size=kernel_size,
            stride=stride,
            padding=pad,
            ceil_mode=ceil_mode,
        )
        self._test_pool(
            context,
            test_input,
            [
                [kernel_size, kernel_size],
                [stride, stride],
                [pad, pad],
                [1, 1,],  # dilation
                ceil_mode,
            ],
            "max_pool2d",
            ops.max_pool2d,
            expected_result,
        )

    @pytest.mark.parametrize(
        "dim, start, end, step",
        itertools.product([0, 1, 2], [0, 1, 2], [3, 4, 5, None], [1, 2]),
    )
    def test_slice(self, context, dim, start, end, step):
        test_input = torch.rand(5, 5, 5)
        constants, input_list, output_name = self._gen_constants(
            5, [test_input, dim, start, end, step]
        )
        node = InternalTorchIRNode(
            kind="slice", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops._slice, node, output_name, constants=constants
        )
        if end is None:
            end = test_input.shape[dim]
        expected_result = test_input.index_select(
            dim, torch.LongTensor(range(start, end, step))
        )
        np.testing.assert_allclose(expected_result, ssa.val)

    @pytest.mark.parametrize(
        "split_sizes, dim, make_explicit",
        itertools.product([2, 3], [0, 1, 2], [True, False]),
    )
    def test_split(self, context, split_sizes, dim, make_explicit):
        test_input = torch.rand(3, 4, 5)
        if make_explicit:
            # Explicitly provide the size of each split. This will be two
            # splits, the given size and the remainder.
            split_sizes = [split_sizes, test_input.shape[dim] - split_sizes]
        constants, input_list, output_name = self._gen_constants(
            3, [test_input, split_sizes, dim]
        )
        node = InternalTorchIRNode(
            kind="split", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.split, node, output_name, constants=constants
        )
        expected_result = torch.split(test_input, split_sizes, dim)
        if not isinstance(ssa, list):
            ssa = [ssa]

        for ex_res, ssa_res in zip(expected_result, ssa):
            np.testing.assert_allclose(ex_res.numpy(), ssa_res.val, atol=1e-6)

    def test_floor(self, context):
        test_input = torch.rand(1, 2, 3) * 10
        constants, input_list, output_name = self._gen_constants(1, test_input)
        floor_node = InternalTorchIRNode(
            kind="floor", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.floor, floor_node, output_name, constants=constants,
        )
        expected_result = test_input.floor()
        assert np.allclose(expected_result, ssa.val)

    def test_erf(self, context):
        test_input = torch.rand(1, 2, 3, 4)
        constants, input_list, output_name = self._gen_constants(1, test_input)
        node = InternalTorchIRNode(kind="erf", inputs=input_list, outputs=[output_name])
        ssa = self._construct_test_graph(
            context, ops.erf, node, output_name, constants=constants
        )
        expected_result = test_input.erf()
        assert np.allclose(expected_result, ssa.val)

    def test_implicittensortonum(self, context):
        input_shape = (1,)
        graph_input_name = "input1"
        graph_inputs = {
            graph_input_name: mb.placeholder(input_shape, dtype=types.float)
        }
        output_name = "1"
        node = InternalTorchIRNode(
            kind="implicittensortonum", inputs=["input1"], outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context,
            ops.implicittensortonum,
            node,
            output_name,
            graph_inputs=graph_inputs,
        )
        assert ssa.shape == ()

    @pytest.mark.parametrize(
        "chunks, dim", itertools.product([2, 3, 5], [0, 1, 2, 3]),
    )
    def test_constantchunk(self, context, chunks, dim):
        test_input = torch.rand(5, 8, 9, 11)
        expected_result = test_input.chunk(chunks, dim=dim)
        constants, input_list, first_output = self._gen_constants(1, [test_input])
        outputs = [str(int(first_output) + i) for i in range(len(expected_result))]
        node = InternalTorchIRNode(
            attr={"chunks": chunks, "dim": dim},
            kind="constantchunk",
            inputs=input_list,
            outputs=outputs,
        )
        self._construct_test_graph(
            context, ops.constantchunk, node, first_output, constants=constants
        )
        actual_result = [context[name] for name in outputs]

        np.testing.assert_equal(len(expected_result), len(actual_result))
        for ex_res, ssa_res in zip(expected_result, actual_result):
            np.testing.assert_allclose(ex_res.numpy(), ssa_res.val, atol=1e-6)

    @pytest.mark.parametrize(
        "input_shape, shape",
        [
            ((3, 1), (3, 4)),
            ((3, 1), (-1, 4)),
            ((3, 1, 1), (3, 4, 1)),
            ((3, 1, 1), (3, -1, 5)),
            ((3, 1, 1), (3, 4, 5)),
            ((1, 3, 1, 1), (2, 3, -1, 1)),
            ((1, 3, 4, 1), (2, 3, -1, 5)),
        ],
    )
    def test_expand(self, context, input_shape, shape):
        test_input = torch.rand(input_shape)
        constants, input_list, output_name = self._gen_constants(2, [test_input, shape])
        node = InternalTorchIRNode(
            kind="expand", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.expand, node, output_name, constants=constants
        )
        expected_result = test_input.expand(shape)
        np.testing.assert_allclose(expected_result, ssa.val)

    @pytest.mark.parametrize(
        "input_shape, other_shape",
        [
            ((3, 1), (3, 4)),
            ((3, 1, 1), (3, 4, 1)),
            ((3, 1, 1), (3, 4, 5)),
            ((1, 3, 1, 1), (2, 3, 4, 1)),
            ((1, 3, 4, 1), (2, 3, 4, 5)),
            ((1, 3, 4, 1), (1, 3, 4, 5)),
        ],
    )
    def test_expand_as(self, context, input_shape, other_shape):
        test_input = torch.rand(input_shape)
        other = torch.rand(other_shape)
        constants, input_list, output_name = self._gen_constants(2, [test_input, other])
        node = InternalTorchIRNode(
            kind="expand_as", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.expand_as, node, output_name, constants=constants
        )
        expected_result = test_input.expand_as(other)
        np.testing.assert_allclose(expected_result, ssa.val)

    @pytest.mark.parametrize(
        "start, end, step",
        [x for x in itertools.product((None, 0, 2), (5, 10), (None,),)]
        + [x for x in itertools.product((0, 2), (5, 10), (1, 2))],
    )
    def test_arange(self, context, start, end, step):
        # Arange can get [end], [start, end], or [start, end, step]
        args = [x for x in [start, end, step] if x is not None]
        args += [0, 0, 0, False]  # Extra args needed but ignored by arange
        constants, input_list, output_name = self._gen_constants(len(args), args)
        node = InternalTorchIRNode(
            kind="arange", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.arange, node, output_name, constants=constants
        )
        kwargs = {"end": end}
        if start is not None:
            kwargs["start"] = start
        if step is not None:
            kwargs["step"] = step
        expected_result = torch.arange(**kwargs)
        np.testing.assert_allclose(expected_result, ssa.val)

    @pytest.mark.parametrize(
        "input_shape, axis",
        [((2, 3), 0), ((2, 3, 4), 1), ((2, 3, 4, 5), 0), ((2, 3, 4, 5), 2),],
    )
    def test_masked_fill(self, context, input_shape, axis):
        mask_shape = list(input_shape)
        mask_shape[axis] = 1
        mask = torch.randint(0, 1, mask_shape, dtype=torch.bool)
        input_data = torch.rand(input_shape)
        value = -1.0
        constants, input_list, output_name = self._gen_constants(
            3, [input_data, mask, value]
        )
        node = InternalTorchIRNode(
            kind="masked_fill", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.masked_fill, node, output_name, constants=constants
        )
        expected_result = input_data.masked_fill(mask, value)
        np.testing.assert_allclose(expected_result, ssa.val)

    @pytest.mark.parametrize("sizes", itertools.permutations([1, 2, 3]))
    def test_meshgrid(self, context, sizes):
        input_tensors = [torch.rand(size) for size in sizes]
        expected_results = torch.meshgrid(input_tensors)
        constants, input_list, output_name = self._gen_constants(3, input_tensors)
        node = InternalTorchIRNode(
            kind="meshgrid", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.meshgrid, node, output_name, constants=constants,
        )
        for expected_result, ssa_result in zip(expected_results, ssa):
            np.testing.assert_allclose(expected_result.numpy(), ssa_result.val)

    @pytest.mark.parametrize(
        "noop_kind",
        ["dropout", "dropout_", "feature_dropout", "contiguous", "device", "detach"],
    )
    def test_noops(self, context, noop_kind):
        test_input = torch.rand(3, 4, 5)
        constants, input_list, output_name = self._gen_constants(
            3, [test_input, "test", "test"]
        )
        node = InternalTorchIRNode(
            kind=noop_kind, inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.noop, node, output_name, constants=constants
        )
        assert np.allclose(test_input.numpy(), ssa.val)

    def test_tanh(self, context):
        test_input = torch.rand(3, 4, 5)
        constants, input_list, output_name = self._gen_constants(1, [test_input])
        node = InternalTorchIRNode(
            kind="tanh", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.tanh, node, output_name, constants=constants
        )
        expected_result = torch.tanh(test_input)
        assert np.allclose(expected_result.numpy(), ssa.val)

    # TODO: test for @keepdim==True when the backend bug is fixed.
    # rdar://62566799
    @pytest.mark.parametrize(
        "input_shape, dim, keepdim",
        itertools.product([(3, 20, 20), (1, 50, 50)], [0, 1, 2], [False]),
    )
    def test_argmax(self, context, input_shape, dim, keepdim):
        test_input = torch.rand(*input_shape)

        constants, input_list, output_name = self._gen_constants(
            4, [test_input, dim, keepdim, None]
        )
        node = InternalTorchIRNode(
            kind="argmax", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.argmax, node, output_name, constants=constants
        )
        expected_result = torch.argmax(test_input, dim, keepdim)
        np.testing.assert_allclose(expected_result, ssa.val)

    @pytest.mark.parametrize(
        "size, dtype", itertools.product([(1, 2, 3, 4), (1,)], [11, 0, 1, 6]),
    )
    def test_zeros(self, context, size, dtype):
        layout = 0  # unused
        device = 0  # unused
        pin_memory = 0  # unused
        constants, input_list, output_name = self._gen_constants(
            5, [size, dtype, layout, device, pin_memory]
        )
        node = InternalTorchIRNode(
            kind="zeros", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.zeros, node, output_name, constants=constants
        )
        expected_result = torch.zeros(size, dtype=ops.NUM_TO_TORCH_DTYPE[dtype])
        np.testing.assert_allclose(expected_result, ssa.val)

    # TODO: Reduce rtol
    # rdar://62868763 (Numerical discrepancy between torch.exp and coreml MIL exp operation)
    @pytest.mark.parametrize("input_size", [(1, 2, 3, 4), (1,)])
    def test_exp(self, context, input_size):
        test_input = torch.rand(input_size)
        constants, input_list, output_name = self._gen_constants(1, test_input)
        node = InternalTorchIRNode(kind="exp", inputs=input_list, outputs=[output_name])
        ssa = self._construct_test_graph(
            context, ops.exp, node, output_name, constants=constants
        )
        expected_result = torch.exp(test_input)
        np.testing.assert_allclose(expected_result, ssa.val, rtol=1e-06)

    @pytest.mark.parametrize(
        "input_size, dim, keepdim",
        itertools.product([(1, 2, 3, 4)], [0, 1, 2], [True, False]),
    )
    def test_max(self, context, input_size, dim, keepdim):
        test_input = torch.rand(input_size)
        constants, input_list, _ = self._gen_constants(3, [test_input, dim, keepdim])
        node = InternalTorchIRNode(
            kind="max", inputs=input_list, outputs=["out1", "out2"],
        )
        ssa = self._construct_test_graph(context, ops.max, node, constants=constants)
        index_result = context["out1"].val
        max_result = context["out2"].val
        expected_index, expected_max = torch.max(test_input, dim=dim, keepdim=keepdim)

    @pytest.mark.parametrize(
        "input_size, dim, descending",
        itertools.product([(2, 3, 4), (1, 2, 3, 4)], [0, 1, 2], [True, False]),
    )
    def test_sort(self, context, input_size, dim, descending):
        test_input = torch.rand(input_size)
        constants, input_list, output_name = self._gen_constants(
            3, [test_input, dim, descending]
        )
        node = InternalTorchIRNode(
            kind="sort", inputs=input_list, outputs=["out1", "out2"],
        )
        ssa = self._construct_test_graph(context, ops.sort, node, constants=constants)
        expected_sort, expected_index = torch.sort(
            test_input, dim=dim, descending=descending
        )
        sort_result = context["out1"].val
        index_result = context["out2"].val
        np.testing.assert_allclose(expected_sort, sort_result)
        np.testing.assert_allclose(expected_index, index_result)

    @pytest.mark.parametrize(
         "input_shape, dim, keepdim",
         itertools.product(
             [(3, 20, 20), (1, 50, 50)],
             [[0], [1], [2], [0, 2]],
             [True, False]),
     )
    def test_sum(self, context, input_shape, dim, keepdim):
        test_input = torch.rand(*input_shape)

        constants, input_list, output_name = self._gen_constants(
            4, [test_input, dim, keepdim, None]
        )
        sum_node = InternalTorchIRNode(
            kind="sum", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.mean, sum_node, output_name, constants=constants
        )
        expected_result = torch.sum(test_input, dim, keepdim)
        assert np.allclose(expected_result, ssa.val)

    def test_sum_no_dims(self, context):
        test_input = torch.rand((3, 20, 20))

        constants, input_list, output_name = self._gen_constants(2, [test_input, None])
        sum_node = InternalTorchIRNode(
            kind="sum", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.mean, sum_node, output_name, constants=constants
        )
        expected_result = torch.sum(test_input)
        assert np.allclose(expected_result, ssa.val)

    def test_neg(self, context):
        test_input = torch.rand(3, 4, 5)
        constants, input_list, output_name = self._gen_constants(1, [test_input])
        node = InternalTorchIRNode(
            kind="neg", inputs=input_list, outputs=[output_name]
        )
        ssa = self._construct_test_graph(
            context, ops.neg, node, output_name, constants=constants
        )
        expected_result = torch.neg(test_input)
        assert np.allclose(expected_result.numpy(), ssa.val)

    @pytest.mark.parametrize(
        "input_shape, k, dim, largest",
        itertools.product([(5, 10, 10), (10, 5, 5)], [0, 3, 5], [0, 1, 2], [True, False]),
    )
    def test_topk(self, context, input_shape, k, dim, largest):
        test_input = torch.tensor(random_gen(input_shape, allow_duplicate=False))

        constants, input_list, output_name = self._gen_constants(
            6, [test_input, k, dim, largest, True, None]
        )
        topk_node = InternalTorchIRNode(
            kind="topk", inputs=input_list, outputs=["out1", "out2"]
        )
        ssa = self._construct_test_graph(
            context, ops.topk, topk_node, constants=constants
        )
        topk_result = context["out1"].val
        index_result = context["out2"].val

        expected_max, expected_indices = torch.topk(test_input, k, dim, largest)
        np.testing.assert_allclose(expected_max.numpy(), topk_result)
        np.testing.assert_allclose(expected_indices.numpy(), index_result)



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil.mil import Builder as mb
from coremltools.converters.mil.testing_utils import (
    assert_op_count_match,
    assert_model_is_valid,
    get_op_types_in_program,
    apply_pass_and_basic_check,
)
import unittest
import pytest

import numpy as np

np.random.seed(1984)


class TransposeOptimizationPass(unittest.TestCase):
    """"""

    """
    Input graph:
    input -----> transpose(axis=[1,0]) -----> transpose(axis=[1,0]) ---> out

    Output graph:
    input -----> relu -----> out
    """

    def test_simple_consecutive_ops_fusion_direct_output(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(10, 20))])
        def prog(x):
            x = mb.transpose(x=x, perm=[1, 0])
            x = mb.transpose(x=x, perm=[1, 0])
            return x

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog), ["transpose", "transpose"]
        )
        self.assertEqual(get_op_types_in_program(prog), ["identity"])
        assert_model_is_valid(
            prog,
            {"x": (10, 20)},
            expected_output_shapes={block.outputs[0].name: (10, 20)},
        )

    """
    Input graph:
    input -----> transpose(axis=[1,0]) -----> transpose(axis=[1,0]) ----> relu ---> out

    Output graph:
    input -----> relu -----> out
    """

    def test_simple_consecutive_ops_fusion(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(10, 20))])
        def prog(x):
            x = mb.transpose(x=x, perm=[1, 0])
            x = mb.transpose(x=x, perm=[1, 0])
            x = mb.relu(x=x)
            return x

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog), ["transpose", "transpose", "relu"]
        )
        self.assertEqual(get_op_types_in_program(prog), ["relu"])
        assert_model_is_valid(
            prog,
            {"x": (10, 20)},
            expected_output_shapes={block.outputs[0].name: (10, 20)},
        )

    """
    Input graph:
    input---->transpose(axis=[0,3,1,2])---->relu---->log--->transpose(axis=[0,2,3,1])--->relu--->out

    Output graph:
    input----->relu----->log----->relu--->out
    """

    def test_linear_graph_two_op_fusion(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 2, 3, 4))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 3, 1, 2])
            x = mb.relu(x=x)
            x = mb.log(x=x)
            x = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x = mb.relu(x=x)
            return x

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["transpose", "relu", "log", "transpose", "relu"],
        )
        self.assertEqual(get_op_types_in_program(prog), ["relu", "log", "relu"])
        assert_model_is_valid(
            prog,
            {"x": (1, 2, 3, 4)},
            expected_output_shapes={block.outputs[0].name: (1, 2, 3, 4)},
        )

    """
    Input graph:
    input---->transpose(axis=[0,3,1,2])---->relu---->identity--->transpose(axis=[0,2,3,1])--->relu--->out

    Output graph:
    input----->relu----->identity----->relu--->out
    """

    def test_linear_graph_two_op_fusion_1(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 2, 3, 4))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 3, 1, 2])
            x = mb.relu(x=x)
            x = mb.identity(x=x)
            x = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x = mb.relu(x=x)
            return x

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["transpose", "relu", "identity", "transpose", "relu"],
        )
        self.assertEqual(get_op_types_in_program(prog), ["relu", "identity", "relu"])
        assert_model_is_valid(
            prog,
            {"x": (1, 2, 3, 4)},
            expected_output_shapes={block.outputs[0].name: (1, 2, 3, 4)},
        )

    """
    Input graph:
    input(shape=1,2,3,4)---->transpose(axis=[0,3,1,2])---->relu---->log--->transpose(axis=[0,2,3,1])--->relu--->out1(shape=1,2,3,4)
                                                                |
                                                                v
                                                        out2(shape=1,4,2,3)

    Output graph:
    input(shape=1,2,3,4)---->relu---->log--->relu--->out1(shape=1,2,3,4)
                                  |
                                  |----->transpose(axis=[0,3,1,2])----->out2(shape=1,4,2,3)
    """

    def test_fusion_with_output_edge_inbetween(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 2, 3, 4))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 3, 1, 2])
            x1 = mb.relu(x=x)
            x2 = mb.log(x=x1)
            x3 = mb.transpose(x=x2, perm=[0, 2, 3, 1])
            x4 = mb.relu(x=x3)
            return x4, x1

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["transpose", "relu", "log", "transpose", "relu"],
        )
        self.assertEqual(
            get_op_types_in_program(prog), ["relu", "log", "relu", "transpose"]
        )
        assert_model_is_valid(
            prog,
            {"x": (1, 2, 3, 4)},
            expected_output_shapes={
                block.outputs[0].name: (1, 2, 3, 4),
                block.outputs[1].name: (1, 4, 2, 3),
            },
        )

    """
    Input graph:
    input---->transpose(axis=[0,3,1,2])---->relu---->transpose(axis=[0,2,3,1])--->out

    Output graph:
    input----->relu----->out
    """

    def test_linear_graph_two_op_fusion_with_last_op_removal(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 2, 3, 4))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 3, 1, 2])
            x = mb.relu(x=x)
            x = mb.transpose(x=x, perm=[0, 2, 3, 1])
            return x

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog), ["transpose", "relu", "transpose"]
        )
        self.assertEqual(get_op_types_in_program(prog), ["relu"])
        assert_model_is_valid(
            prog,
            {"x": (1, 2, 3, 4)},
            expected_output_shapes={block.outputs[0].name: (1, 2, 3, 4)},
        )

    """
    Input graph:
    input(shape=10,2,3)--->transpose(axis=[0,2,1])----->relu---->transpose(axis=[0,2,1])---->out1
                                                    |
                                                    |
                                                    --->relu----->log---->transpose(axis=[0,2,1])---->out2

    Output graph:
    input(shape=10,2,3)----->relu---->out1
                        |
                        |
                        --->relu----->log---->out2
    """

    def test_multiple_fusions(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(10, 2, 3))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 2, 1])
            x1 = mb.relu(x=x)
            x2 = mb.relu(x=x)
            y1 = mb.transpose(x=x1, perm=[0, 2, 1])
            x3 = mb.log(x=x2)
            y2 = mb.transpose(x=x3, perm=[0, 2, 1])
            return y1, y2

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["transpose", "relu", "relu", "transpose", "log", "transpose"],
        )
        self.assertEqual(get_op_types_in_program(prog), ["relu", "relu", "log"])

        assert (
            prev_block.inputs["x"]
            == prev_block.find_ops(op_type="transpose")[0].inputs["x"]
        )
        assert block.find_ops(op_type="log")[0].outputs[0] in block.outputs
        assert_model_is_valid(
            prog,
            {"x": (10, 2, 3)},
            expected_output_shapes={
                block.outputs[0].name: (10, 2, 3),
                block.outputs[1].name: (10, 2, 3),
            },
        )

    """
    Input graph:
    input(shape=10,2,3,5)--->transpose(axis=[0,2,3,1])----->relu---->pool----->out1
                                                       |
                                                       |
                                                       --->relu----->log---->transpose(axis=[0,3,1,2])---->out2


    Output graph:
    input(shape=10,2,3,5)----->relu---->transpose(axis=[0,2,3,1])---->pool----->out1
                           |
                           |
                           --->relu----->log---->out2
    """

    def test_partial_fusion_0(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(10, 2, 3, 5))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x1 = mb.relu(x=x)
            x2 = mb.relu(x=x)
            y1 = mb.avg_pool(
                x=x1, kernel_sizes=[1, 1], strides=[1, 1], pad_type="valid"
            )
            x3 = mb.log(x=x2)
            y2 = mb.transpose(x=x3, perm=[0, 3, 1, 2])
            return y1, y2

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["transpose", "relu", "relu", "avg_pool", "log", "transpose"],
        )
        self.assertEqual(
            get_op_types_in_program(prog),
            ["relu", "relu", "transpose", "avg_pool", "log"],
        )

        assert (
            prev_block.inputs["x"]
            == prev_block.find_ops(op_type="transpose")[0].inputs["x"]
        )
        assert block.find_ops(op_type="log")[0].outputs[0] == block.outputs[1]
        assert (
            block.find_ops(op_type="transpose")[0].outputs[0]
            == block.find_ops(op_type="avg_pool")[0].inputs["x"]
        )
        assert list(block.find_ops(op_type="transpose")[0].perm.val) == [0, 2, 3, 1]
        assert_model_is_valid(
            prog,
            {"x": (10, 2, 3, 5)},
            expected_output_shapes={
                block.outputs[0].name: (10, 3, 5, 2),
                block.outputs[1].name: (10, 2, 3, 5),
            },
        )

    """
    Input graph:
    input(shape=10,2,3,5)--->transpose(axis=[0,2,1,3])----->relu---->transpose(axis=[0,2,1,3])---->out1
                                                        |
                                                        |
                                                        --->pool--->log---->transpose(axis=[0,2,1,3])---->out2

    Output graph:
    input(shape=10,2,3,5)----->relu---->out1
                           |
                           |
                           --->transpose(axis=[0,2,1,3])---->pool----->log---->transpose(axis=[0,2,1,3])---->out2
    """

    def test_partial_fusion_1(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(10, 2, 3, 5))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 2, 1, 3])
            x1 = mb.relu(x=x)
            x2 = mb.avg_pool(x=x, kernel_sizes=[1, 1], strides=[1, 1], pad_type="valid")
            y1 = mb.transpose(x=x1, perm=[0, 2, 1, 3])
            x3 = mb.log(x=x2)
            y2 = mb.transpose(x=x3, perm=[0, 2, 1, 3])
            return y1, y2

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["transpose", "relu", "avg_pool", "transpose", "log", "transpose"],
        )
        self.assertEqual(
            get_op_types_in_program(prog),
            ["relu", "transpose", "avg_pool", "log", "transpose"],
        )

        assert block.inputs["x"] == block.find_ops(op_type="relu")[0].inputs["x"]
        assert block.outputs[0] == block.find_ops(op_type="relu")[0].outputs[0]
        assert_model_is_valid(
            prog,
            {"x": (10, 2, 3, 5)},
            expected_output_shapes={
                block.outputs[0].name: (10, 2, 3, 5),
                block.outputs[1].name: (10, 2, 3, 5),
            },
        )

    """
    Input graph:

                                                             |-------> transpose(axis=[0,2,1,3]) ---->out1(shape=10,2,3,5)
                                                             |
    input(shape=10,2,3,5)-->relu-->transpose(axis=[0,2,1,3])--->relu--->transpose(axis=[0,2,1,3]) ---->out2(shape=10,2,3,5)
                                                                     |
                                                                     |----->pool--------------->out3(shape=10,3,2,5)
                                                                     |
                                                                     |----->pool--------------->out4(shape=10.3.2.5)


    Output graph:

                        |---->out1(shape=10,2,3,5)
                        |
    input---->relu---------->relu------->out2(shape=10,2,3,5)
                                    |
                                    |----->transpose(axis=[0,2,1,3])--->pool---->out3(shape=10,3,2,5)
                                    |
                                    |----->transpose(axis=[0,2,1,3])---->pool--->out4(shape=10.3.2.5)
    """

    def test_partial_fusion_2(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(10, 2, 3, 5))])
        def prog(x):
            x = mb.relu(x=x)
            x = mb.transpose(x=x, perm=[0, 2, 1, 3])
            y1 = mb.transpose(x=x, perm=[0, 2, 1, 3])
            x1 = mb.relu(x=x)
            y2 = mb.transpose(x=x1, perm=[0, 2, 1, 3])
            y3 = mb.avg_pool(
                x=x1, kernel_sizes=[1, 1], strides=[1, 1], pad_type="valid"
            )
            y4 = mb.avg_pool(
                x=x1, kernel_sizes=[1, 1], strides=[1, 1], pad_type="valid"
            )
            return y1, y2, y3, y4

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            [
                "relu",
                "transpose",
                "transpose",
                "relu",
                "transpose",
                "avg_pool",
                "avg_pool",
            ],
        )
        self.assertEqual(
            get_op_types_in_program(prog),
            ["relu", "relu", "transpose", "avg_pool", "transpose", "avg_pool"],
        )

        assert block.outputs[0] == block.find_ops(op_type="relu")[0].outputs[0]
        assert block.outputs[1] == block.find_ops(op_type="relu")[1].outputs[0]
        assert_model_is_valid(
            prog,
            {"x": (10, 2, 3, 5)},
            expected_output_shapes={
                block.outputs[0].name: (10, 2, 3, 5),
                block.outputs[1].name: (10, 2, 3, 5),
                block.outputs[2].name: (10, 3, 2, 5),
                block.outputs[3].name: (10, 3, 2, 5),
            },
        )

    """
    Input graph:

    input(shape=10,2,3,5)-->relu--->transpose(axis=[0,2,1,3])----->transpose(axis=[0,2,1,3])---->out1(shape=10,2,3,5)
                                                               |
                                                               ---->relu------>out2(shape=10,3,2,5)

    Output graph:

    input(shape=10,2,3,5)-->relu---->out1(shape=10,2,3,5)
                                  |
                                   ---->relu--->transpose(axis=[0,2,1,3])------>out2(shape=10,3,2,5)
    """

    def test_partial_fusion_3(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(10, 2, 3, 5))])
        def prog(x):
            x = mb.relu(x=x)
            x = mb.transpose(x=x, perm=[0, 2, 1, 3])
            x1 = mb.transpose(x=x, perm=[0, 2, 1, 3])
            x2 = mb.relu(x=x)
            return x1, x2

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["relu", "transpose", "transpose", "relu"],
        )
        self.assertEqual(get_op_types_in_program(prog), ["relu", "relu", "transpose"])

        assert block.outputs[0] == block.find_ops(op_type="relu")[0].outputs[0]
        assert_model_is_valid(
            prog,
            {"x": (10, 2, 3, 5)},
            expected_output_shapes={
                block.outputs[0].name: (10, 2, 3, 5),
                block.outputs[1].name: (10, 3, 2, 5),
            },
        )

    """
    Input graph:

    input(shape=10,2,3,5)-->relu--->transpose(axis=[0,2,1,3])----->transpose(axis=[0,2,1,3])---->out1(shape=10,2,3,5)
                                                               |
                                                               ------>out2(shape=10,3,2,5)

    Output graph:
    same as input graph as one of the optimizing transpose is connected to model output
    """

    def test_partial_fusion_4(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(10, 2, 3, 5))])
        def prog(x):
            x = mb.relu(x=x)
            out2 = mb.transpose(x=x, perm=[0, 2, 1, 3])
            out1 = mb.transpose(x=out2, perm=[0, 2, 1, 3])
            return out1, out2

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog), ["relu", "transpose", "transpose"]
        )
        self.assertEqual(
            get_op_types_in_program(prog), ["relu", "transpose", "transpose"]
        )

        assert block.outputs[1] == block.find_ops(op_type="transpose")[0].outputs[0]
        assert_model_is_valid(
            prog,
            {"x": (10, 2, 3, 5)},
            expected_output_shapes={
                block.outputs[0].name: (10, 2, 3, 5),
                block.outputs[1].name: (10, 3, 2, 5),
            },
        )

    """
    Input graph:
    input(shape=10,2,3,5)-->relu-->transpose(axis=[0,2,1,3])--->relu--->transpose(axis=[0,2,1,3]) ---->out1(shape=10,2,3,5)
                                                                     |
                                                                     |--->relu-->pool--------------->out2(shape=10,3,2,5)
                                                                     |
                                                                     |----->pool--------------->out3(shape=10.3.2.5)


    Output graph:
    same as the input graph as materialization ops are greater than cancel ops
    """

    def test_no_fusion_more_materialization_ops(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(10, 2, 3, 5))])
        def prog(x):
            x = mb.relu(x=x)
            x = mb.transpose(x=x, perm=[0, 2, 1, 3])
            x1 = mb.relu(x=x)
            y2 = mb.transpose(x=x1, perm=[0, 2, 1, 3])
            x2 = mb.relu(x=x1)
            y3 = mb.avg_pool(
                x=x2, kernel_sizes=[1, 1], strides=[1, 1], pad_type="valid"
            )
            y4 = mb.avg_pool(
                x=x1, kernel_sizes=[1, 1], strides=[1, 1], pad_type="valid"
            )
            return y2, y3, y4

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["relu", "transpose", "relu", "transpose", "relu", "avg_pool", "avg_pool"],
        )
        self.assertEqual(
            get_op_types_in_program(prog),
            ["relu", "transpose", "relu", "transpose", "relu", "avg_pool", "avg_pool"],
        )

        assert_model_is_valid(
            prog,
            {"x": (10, 2, 3, 5)},
            expected_output_shapes={
                block.outputs[0].name: (10, 2, 3, 5),
                block.outputs[1].name: (10, 3, 2, 5),
                block.outputs[2].name: (10, 3, 2, 5),
            },
        )

    """
    Input graph:
    input(shape=10,2,3)--->transpose(axis=[0,2,1])----->relu---->transpose(axis=[0,2,1])---->out1
                                                    |
                                                    |
                                                    --->reduce(axis=2)----->log---->transpose(axis=[0,2,1])---->out2

    Output graph:
    input(shape=10,2,3)----->relu---->out1
                        |
                        |
                        --->reduce(axis=1)----->log---->out2
    """

    def test_fusion_with_axis_op(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(10, 2, 3))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 2, 1])
            x1 = mb.relu(x=x)
            x2 = mb.reduce_mean(x=x, axes=[2], keep_dims=True)
            y1 = mb.transpose(x=x1, perm=[0, 2, 1])
            x3 = mb.log(x=x2)
            y2 = mb.transpose(x=x3, perm=[0, 2, 1])
            return y1, y2

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["transpose", "relu", "reduce_mean", "transpose", "log", "transpose"],
        )
        self.assertEqual(get_op_types_in_program(prog), ["relu", "reduce_mean", "log"])

        assert list(block.find_ops(op_type="reduce_mean")[0].inputs["axes"].val) == [1]
        assert_model_is_valid(
            prog,
            {"x": (10, 2, 3)},
            expected_output_shapes={
                block.outputs[0].name: (10, 2, 3),
                block.outputs[1].name: (10, 1, 3),
            },
        )

    """
    Input graph:
    input(shape=11,2,3,6)--->transpose(axis=[0,3,1,2])---
                                                       |
                                                       |
                                                        --->pad(pad=[0,0,0,0,1,2,3,4])
                                                              |
                                                              |-->log--->transpose(axis=[0,2,3,1])-->out1(shape=11,5,10,6)

    Output graph:
    same as input graph, as transpose cannot be pushed through the pad op since "reflect" mode is only supported
    along the last two axis
    """

    def test_fusion_with_pad_reflective_op_0(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(11, 2, 3, 6))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 3, 1, 2])
            x2 = mb.pad(x=x, pad=[0, 0, 0, 0, 1, 2, 3, 4], mode="reflect")
            x3 = mb.log(x=x2)
            y2 = mb.transpose(x=x3, perm=[0, 2, 3, 1])
            return y2

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog), ["transpose", "pad", "log", "transpose"]
        )
        self.assertEqual(
            get_op_types_in_program(prog), ["transpose", "pad", "log", "transpose"]
        )

        assert list(block.find_ops(op_type="pad")[0].inputs["pad"].val.flatten()) == [
            0,
            0,
            0,
            0,
            1,
            2,
            3,
            4,
        ]
        assert_model_is_valid(
            prog,
            {"x": (11, 2, 3, 6)},
            expected_output_shapes={block.outputs[0].name: (11, 5, 10, 6)},
        )

    """
    Input graph:
    input(shape=11,2,3,6)--->transpose(axis=[0,1,3,2])---
                                                       |
                                                       |
                                                        --->pad(pad=[0,0,0,0,1,2,3,4])
                                                              |
                                                              |-->log--->transpose(axis=[0,1,3,2])-->out1(shape=11,2,10,9)

    Output graph:
    input(shape=11,2,3,6)--->pad(pad=[0,0,0,0,3,4,1,2])-->log-->out1(shape=11,2,10,9)
    """

    def test_fusion_with_pad_reflective_op_1(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(11, 2, 3, 6))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 1, 3, 2])
            x2 = mb.pad(x=x, pad=[0, 0, 0, 0, 1, 2, 3, 4], mode="reflect")
            x3 = mb.log(x=x2)
            y2 = mb.transpose(x=x3, perm=[0, 1, 3, 2])
            return y2

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog), ["transpose", "pad", "log", "transpose"]
        )
        self.assertEqual(get_op_types_in_program(prog), ["pad", "log"])

        assert list(block.find_ops(op_type="pad")[0].inputs["pad"].val.flatten()) == [
            0,
            0,
            0,
            0,
            3,
            4,
            1,
            2,
        ]
        assert_model_is_valid(
            prog,
            {"x": (11, 2, 3, 6)},
            expected_output_shapes={block.outputs[0].name: (11, 2, 10, 9)},
        )

    """
    Input graph:
    input(shape=11,2,3,6)--->transpose(axis=[0,3,1,2])---
                                                       |
                                                       |
                                                        --->pad(pad=[0,0,0,0,1,2,3,4])
                                                              |
                                                              |-->log--->transpose(axis=[0,2,3,1])-->out1(shape=11,5,10,6)

    Output graph:
    input(shape=11,2,3,6)--->pad(pad=[0,0,1,2,3,4,0,0])-->log-->out1(shape=11,5,10,6)
    """

    def test_fusion_with_pad_constant_op(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(11, 2, 3, 6))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 3, 1, 2])
            x2 = mb.pad(
                x=x, pad=[0, 0, 0, 0, 1, 2, 3, 4], mode="constant", constant_val=3.0
            )
            x3 = mb.log(x=x2)
            y2 = mb.transpose(x=x3, perm=[0, 2, 3, 1])
            return y2

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog), ["transpose", "pad", "log", "transpose"]
        )
        self.assertEqual(get_op_types_in_program(prog), ["pad", "log"])

        assert list(block.find_ops(op_type="pad")[0].inputs["pad"].val.flatten()) == [
            0,
            0,
            1,
            2,
            3,
            4,
            0,
            0,
        ]
        assert_model_is_valid(
            prog,
            {"x": (11, 2, 3, 6)},
            expected_output_shapes={block.outputs[0].name: (11, 5, 10, 6)},
        )

    """
    Input graph:
                                                    const(shape=2)
                                                          |
                                                          V
    input(shape=1,2,5,5)--->transpose(axis=[0,2,3,1])--->add---->transpose(axis=[0,3,1,2])--->out(shape=1,2,5,5)

    Output graph:
                        const(shape=1,2,1,1)
                             |
                             V
    input(shape=1,2,5,5)--->add--->out(shape=1,2,5,5)
    """

    def test_fusion_with_add_constant_op(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 2, 5, 5))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x = mb.add(x=x, y=np.array([10, 100]))
            x = mb.transpose(x=x, perm=[0, 3, 1, 2])
            return x

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog), ["transpose", "add", "transpose"]
        )
        self.assertEqual(get_op_types_in_program(prog), ["add"])

        assert_model_is_valid(
            prog,
            {"x": (1, 2, 5, 5)},
            expected_output_shapes={block.outputs[0].name: (1, 2, 5, 5)},
        )

    """
    Input graph:
                                                    const(scalar)
                                                          |
                                                          V
    input(shape=1,2,5,5)--->transpose(axis=[0,2,3,1])--->add---->transpose(axis=[0,3,1,2])--->out(shape=1,2,5,5)

    Output graph:
                        const(scalar)
                             |
                             V
    input(shape=1,2,5,5)--->add--->out(shape=1,2,5,5)
    """

    def test_fusion_with_add_scalar_constant_op(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 2, 5, 5))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x = mb.add(x=5, y=x)
            x = mb.transpose(x=x, perm=[0, 3, 1, 2])
            return x

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog), ["transpose", "add", "transpose"]
        )
        self.assertEqual(get_op_types_in_program(prog), ["add"])

        assert_model_is_valid(
            prog,
            {"x": (1, 2, 5, 5)},
            expected_output_shapes={block.outputs[0].name: (1, 2, 5, 5)},
        )

    """
    Input graph:
    input(shape=1,2,5,5)----->transpose(axis=[0,2,3,1])--->add---->transpose(axis=[0,3,1,2])--->out(shape=1,2,5,5)
                          |                                 ^
                          |                                 |
                          |---->relu---->transpose(axis=[0,2,3,1])

    Output graph:
    input(shape=1,2,5,5)----->add--->out(shape=1,2,5,5)
                      |        ^
                      |        |
                      |------>relu
    """

    def test_fusion_with_add_broadcastable_0(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 2, 5, 5))])
        def prog(x):
            x1 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x2 = mb.relu(x=x)
            x2 = mb.transpose(x=x2, perm=[0, 2, 3, 1])
            x3 = mb.add(x=x1, y=x2)
            y = mb.transpose(x=x3, perm=[0, 3, 1, 2])
            return y

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["transpose", "relu", "transpose", "add", "transpose"],
        )
        self.assertEqual(get_op_types_in_program(prog), ["relu", "add"])

        assert block.find_ops(op_type="relu")[0].inputs["x"] == block.inputs["x"]
        assert block.find_ops(op_type="add")[0].inputs["x"] == block.inputs["x"]
        assert (
            block.find_ops(op_type="add")[0].inputs["y"]
            == block.find_ops(op_type="relu")[0].outputs[0]
        )

        assert_model_is_valid(
            prog,
            {"x": (1, 2, 5, 5)},
            expected_output_shapes={block.outputs[0].name: (1, 2, 5, 5)},
        )

    """
    Input graph:
    input(shape=1,2,5,5)----->transpose(axis=[0,2,3,1])--->add---->transpose(axis=[0,3,1,2])--->out(shape=1,2,5,5)
                          |                                 ^
                          |                                 |
                          |----------------------->transpose(axis=[0,2,3,1])

    Output graph:
    input(shape=1,2,5,5)----->add--->out(shape=1,2,5,5)
                      |        ^
                      |        |
                      |---------
    """

    def test_fusion_with_add_broadcastable_1(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 2, 5, 5))])
        def prog(x):
            x1 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x2 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x3 = mb.add(x=x1, y=x2)
            y = mb.transpose(x=x3, perm=[0, 3, 1, 2])
            return y

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["transpose", "transpose", "add", "transpose"],
        )
        self.assertEqual(get_op_types_in_program(prog), ["add"])

        assert block.find_ops(op_type="add")[0].inputs["x"] == block.inputs["x"]
        assert block.find_ops(op_type="add")[0].inputs["y"] == block.inputs["x"]

        assert_model_is_valid(
            prog,
            {"x": (1, 2, 5, 5)},
            expected_output_shapes={block.outputs[0].name: (1, 2, 5, 5)},
        )

    """
    Input graph:
    input(shape=1,2,5,5)--->transpose(axis=[0,2,3,1])---> relu---->concat(axis=3)----->transpose(axis=[0,3,1,2])----->out1(shape=1,4,5,5)
                         |                                              ^
                         |                                              |
                         |->transpose(axis=[0,2,3,1])--->relu------------

    Output graph:
    input(shape=1,2,5,5)------> relu---->concat(axis=1)--->out1(shape=1,4,5,5)
                         |                    ^
                         |                    |
                         |---->relu------------
    """

    def test_concat_pattern_0(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 2, 5, 5))])
        def prog(x):
            x1 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x2 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x1 = mb.relu(x=x1)
            x2 = mb.relu(x=x2)
            x3 = mb.concat(values=[x1, x2], axis=3)
            x4 = mb.transpose(x=x3, perm=[0, 3, 1, 2])
            return x4

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["transpose", "transpose", "relu", "relu", "concat", "transpose"],
        )
        self.assertEqual(get_op_types_in_program(prog), ["relu", "relu", "concat"])

        assert_model_is_valid(
            prog,
            {"x": (1, 2, 5, 5)},
            expected_output_shapes={block.outputs[0].name: (1, 4, 5, 5)},
        )

    """
    Input graph:
    input(shape=1,2,5,5)--->transpose(axis=[0,2,3,1])---> relu---->concat(axis=3)----->transpose(axis=[0,3,1,2])----->out1(shape=1,4,5,5)
                         |                                              ^
                         |                                              |
                         |->transpose(axis=[0,2,3,1])------->relu--------
                                                        |
                                                        V
                                                       pool--->out2(shape=1,5,5,2)



    Output graph:
    input(shape=1,2,5,5)------> relu---->concat(axis=1)--->out1(shape=1,4,5,5)
                         |                    ^
                         |                    |
                         |---->relu------------
                         |
                         |--->transpose(axis=[0,2,3,1])---->pool--->out2(shape=1,5,5,2)
    """

    def test_concat_pattern_1(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 2, 5, 5))])
        def prog(x):
            x1 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x2 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x1 = mb.relu(x=x1)
            x2 = mb.relu(x=x2)
            x3 = mb.concat(values=[x1, x2], axis=3)
            x4 = mb.transpose(x=x3, perm=[0, 3, 1, 2])
            x5 = mb.avg_pool(
                x=x2, kernel_sizes=[1, 1], strides=[1, 1], pad_type="valid"
            )
            return x4, x5

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            [
                "transpose",
                "transpose",
                "relu",
                "relu",
                "concat",
                "transpose",
                "avg_pool",
            ],
        )
        self.assertEqual(
            get_op_types_in_program(prog),
            ["relu", "relu", "concat", "transpose", "avg_pool"],
        )
        assert_model_is_valid(
            prog,
            {"x": (1, 2, 5, 5)},
            expected_output_shapes={
                block.outputs[0].name: (1, 4, 5, 5),
                block.outputs[1].name: (1, 5, 5, 2),
            },
        )

    """
    Input graph:
    input(shape=1,2,5,5)--->transpose(axis=[0,2,3,1])---> relu---->concat(axis=3)----->transpose(axis=[0,3,1,2])----->out1(shape=1,4,5,5)
                         |                                              ^
                         |                                              |
                         |->transpose(axis=[0,2,3,1])------->relu--------
                                                        |
                                                        V
                                                       relu--->out2(shape=1,5,5,2)



    Output graph:
    input(shape=1,2,5,5)------> relu---->concat(axis=1)--->out1(shape=1,4,5,5)
                         |                    ^
                         |                    |
                         |---->relu------------
                         |
                         |--->relu---->transpose(axis=[0,2,3,1])---->out2(shape=1,5,5,2)
    """

    def test_concat_pattern_2(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 2, 5, 5))])
        def prog(x):
            x1 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x2 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x1 = mb.relu(x=x1)
            x2 = mb.relu(x=x2)
            x3 = mb.concat(values=[x1, x2], axis=3)
            x4 = mb.transpose(x=x3, perm=[0, 3, 1, 2])
            x5 = mb.relu(x=x2)
            return x4, x5

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["transpose", "transpose", "relu", "relu", "concat", "transpose", "relu"],
        )
        self.assertEqual(
            get_op_types_in_program(prog),
            ["relu", "relu", "concat", "relu", "transpose"],
        )

        assert_model_is_valid(
            prog,
            {"x": (1, 2, 5, 5)},
            expected_output_shapes={
                block.outputs[0].name: (1, 4, 5, 5),
                block.outputs[1].name: (1, 5, 5, 2),
            },
        )

    """
    Input graph:
    input(shape=1,2,5,5)--->transpose(axis=[0,2,3,1])---> relu---->concat(axis=3)----->transpose(axis=[0,3,1,2])----->out1(shape=1,4,5,5)
                         |                                              ^
                         |                                              |
                         |->transpose(axis=[0,2,3,1])------->relu--------
                                                        |
                                                        V
                                                       out2(shape=1,5,5,2)



    Output graph:
    input(shape=1,2,5,5)------> relu---->concat(axis=1)--->out1(shape=1,4,5,5)
                         |                    ^
                         |                    |
                         |---->relu------------
                         |
                         |--->transpose(axis=[0,2,3,1])---->out2(shape=1,5,5,2)
    """

    def test_concat_pattern_3(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 2, 5, 5))])
        def prog(x):
            x1 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x2 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x1 = mb.relu(x=x1)
            x2 = mb.relu(x=x2)
            x3 = mb.concat(values=[x1, x2], axis=3)
            x4 = mb.transpose(x=x3, perm=[0, 3, 1, 2])
            return x4, x2

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["transpose", "transpose", "relu", "relu", "concat", "transpose"],
        )
        self.assertEqual(
            get_op_types_in_program(prog), ["relu", "relu", "concat", "transpose"]
        )

        assert_model_is_valid(
            prog,
            {"x": (1, 2, 5, 5)},
            expected_output_shapes={
                block.outputs[0].name: (1, 4, 5, 5),
                block.outputs[1].name: (1, 5, 5, 2),
            },
        )

    """
    Input graph:
    input(shape=1,2,5,5)--->transpose(axis=[0,2,3,1])---> relu---->concat(axis=3)----->transpose(axis=[0,3,1,2])----->out1(shape=1,4,5,5)
                         |                                              ^
                         |                                              |
                         |->transpose(axis=[0,2,3,1])------->relu--------
                                                        |
                                                        V
                                            transpose(axis=[0,3,1,2]) -----> out2(shape=1,2,5,5)

    Output graph:
    input(shape=1,2,5,5)---> relu---->concat(axis=1)----->out1(shape=1,4,5,5)
                     |                     ^
                     |                     |
                     |------------------->relu-------->out2(shape=1,2,5,5)
    """

    def test_concat_pattern_4(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 2, 5, 5))])
        def prog(x):
            x1 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x2 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x1 = mb.relu(x=x1)
            x2 = mb.relu(x=x2)
            x3 = mb.concat(values=[x1, x2], axis=3)
            x4 = mb.transpose(x=x3, perm=[0, 3, 1, 2])
            x5 = mb.transpose(x=x2, perm=[0, 3, 1, 2])
            return x4, x5

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            [
                "transpose",
                "transpose",
                "relu",
                "relu",
                "concat",
                "transpose",
                "transpose",
            ],
        )
        self.assertEqual(get_op_types_in_program(prog), ["relu", "relu", "concat"])

        assert_model_is_valid(
            prog,
            {"x": (1, 2, 5, 5)},
            expected_output_shapes={
                block.outputs[0].name: (1, 4, 5, 5),
                block.outputs[1].name: (1, 2, 5, 5),
            },
        )

    """
    Input graph:
                                                     constant(shape=[30,10,5])
                                                            |
                                                            V
    input(shape=10,20,30)--->transpose(axis=[2,0,1])--->concat(axis=2)----->transpose(axis=[1,2,0])----->out1(shape=10,25,30)

    Output graph:
                        constant(shape=[10,5,30])
                                |
                                V
    input(shape=10,20,30)--->concat(axis=1)----->out1(shape=10,25,30)
    """

    def test_concat_pattern_5(self):
        const = np.random.rand(30, 10, 5)

        @mb.program(input_specs=[mb.TensorSpec(shape=(10, 20, 30))])
        def prog(x):
            x1 = mb.transpose(x=x, perm=[2, 0, 1])
            c = mb.const(val=const)
            x2 = mb.concat(values=[x1, c], axis=2)
            x3 = mb.transpose(x=x2, perm=[1, 2, 0])
            return x3

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog), ["transpose", "concat", "transpose"]
        )
        self.assertEqual(get_op_types_in_program(prog), ["concat"])

        assert_model_is_valid(
            prog,
            {"x": (10, 20, 30)},
            expected_output_shapes={block.outputs[0].name: (10, 25, 30)},
        )

    """
    Input graph:
                                        input2(shape=30,10,20)-----|
                                                                   |
    input(shape=10,20,30)--->transpose(axis=[2,0,1])----->relu-----|----->concat(axis=2)------>out1(shape=90,10,20)
                                                      |            |
                                                      |-->relu-----|
                                                      |
                                                      |-->relu---->transpose(axis=[1,2,0])---->out2(shape=10,20,30)
                                                      |
                                                      |-->relu---->transpose(axis=[1,2,0])---->out3(shape=10,20,30)
                                                      |
                                                      |-->relu---->transpose(axis=[1,2,0])---->out4(shape=10,20,30)

    Output graph:

                                        input2(shape=30,10,20)-----|
                                                                   |
    input(shape=10,20,30)----->relu--->transpose(axis=[2,0,1])-----|----->concat(axis=2)------>out1(shape=90,10,20)
                           |                                       |
                           |-->relu--->transpose(axis=[2,0,1])-----|
                           |
                           |-->relu---->out2(shape=10,20,30)
                           |
                           |-->relu---->out3(shape=10,20,30)
                           |
                           |-->relu---->out4(shape=10,20,30)

    Output graph:
    """

    def test_concat_pattern_6(self):
        @mb.program(
            input_specs=[
                mb.TensorSpec(shape=(10, 20, 30)),
                mb.TensorSpec(shape=(30, 10, 20)),
            ]
        )
        def prog(x, y):
            x1 = mb.transpose(x=x, perm=[2, 0, 1])
            r1 = mb.relu(x=x1)
            r2 = mb.relu(x=x1)
            r3 = mb.relu(x=x1)
            r4 = mb.relu(x=x1)
            r5 = mb.relu(x=x1)

            x2 = mb.concat(values=[r1, r2, y], axis=0)
            x3 = mb.transpose(x=r3, perm=[1, 2, 0])
            x4 = mb.transpose(x=r4, perm=[1, 2, 0])
            x5 = mb.transpose(x=r5, perm=[1, 2, 0])
            return x2, x3, x4, x5

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            [
                "transpose",
                "relu",
                "relu",
                "relu",
                "relu",
                "relu",
                "concat",
                "transpose",
                "transpose",
                "transpose",
            ],
        )
        self.assertEqual(
            get_op_types_in_program(prog),
            [
                "relu",
                "relu",
                "relu",
                "relu",
                "relu",
                "transpose",
                "transpose",
                "concat",
            ],
        )

        assert_model_is_valid(
            prog,
            {"x": (10, 20, 30), "y": (30, 10, 20)},
            expected_output_shapes={
                block.outputs[0].name: (90, 10, 20),
                block.outputs[1].name: (10, 20, 30),
                block.outputs[2].name: (10, 20, 30),
                block.outputs[3].name: (10, 20, 30),
            },
        )

    """
    Input graph:
    input(shape=1,5,5,3)----->transpose(axis=[0,3,1,2])
                                |
                                ---->relu-------------->transpose(axis=[0,2,3,1])
                                             |                  |
                                             |                  V
                                             |                 relu
                                             |                  |
                                             |                  V
                                             |         transpose(axis=[0,3,1,2])
                                             |                  |
                                             |                  V
                                             ----------------> add --------> relu---->pool---->out(shape=1,3,5,5)


    Output graph:


    input(shape=1,5,5,3)---->relu------------------------> relu
                                         |                  |
                                         |                  V
                                         ----------------> add
                                                            |
                                                            V
                                                           relu
                                                            |
                                                            V
                                                           transpose(axis=[0,3,1,2])-->pool---->out(shape=1,3,5,5)

    """

    def test_skip_connection_pattern_0(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 5, 5, 3))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 3, 1, 2])
            x = mb.relu(x=x)
            x1 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x2 = mb.relu(x=x1)
            x3 = mb.transpose(x=x2, perm=[0, 3, 1, 2])
            x4 = mb.add(x=x, y=x3)
            x5 = mb.relu(x=x4)
            x6 = mb.avg_pool(
                x=x5, kernel_sizes=[1, 1], strides=[1, 1], pad_type="valid"
            )
            return x6

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            [
                "transpose",
                "relu",
                "transpose",
                "relu",
                "transpose",
                "add",
                "relu",
                "avg_pool",
            ],
        )
        self.assertEqual(
            get_op_types_in_program(prog),
            ["relu", "relu", "add", "relu", "transpose", "avg_pool"],
        )
        assert_model_is_valid(
            prog,
            {"x": (1, 5, 5, 3)},
            expected_output_shapes={block.outputs[0].name: (1, 3, 5, 5)},
        )

    """
    Input graph:
    input(shape=1,5,5,3)----->transpose(axis=[0,3,1,2])
                                |
                                ---->relu-------------->transpose(axis=[0,2,3,1])
                                             |                  |
                                             |                  V
                                             |                 relu
                                             |                  |
                                             |                  V
                                             |         transpose(axis=[0,3,1,2])
                                             |                  |
                                             |                  V
                                             ----------------> add -->transpose(axis=[0,2,3,1])
                                                                            |
                                                                            V
                                                                          relu---->pool---->out(shape=1,5,5,3)


    Output graph:


    input(shape=1,5,5,3)---->relu------------------------> relu
                                         |                  |
                                         |                  V
                                         ----------------> add
                                                            |
                                                            V
                                                           relu
                                                            |
                                                            V
                                                           pool---->out(shape=1,5,5,3)

    """

    def test_skip_connection_pattern_1(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 5, 5, 3))])
        def prog(x):
            x = mb.transpose(x=x, perm=[0, 3, 1, 2])
            x = mb.relu(x=x)
            x1 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x2 = mb.relu(x=x1)
            x3 = mb.transpose(x=x2, perm=[0, 3, 1, 2])
            x4 = mb.add(x=x, y=x3)
            x4 = mb.transpose(x=x4, perm=[0, 2, 3, 1])
            x5 = mb.relu(x=x4)
            x6 = mb.avg_pool(
                x=x5, kernel_sizes=[1, 1], strides=[1, 1], pad_type="valid"
            )
            return x6

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            [
                "transpose",
                "relu",
                "transpose",
                "relu",
                "transpose",
                "add",
                "transpose",
                "relu",
                "avg_pool",
            ],
        )
        self.assertEqual(
            get_op_types_in_program(prog), ["relu", "relu", "add", "relu", "avg_pool"]
        )
        assert_model_is_valid(
            prog,
            {"x": (1, 5, 5, 3)},
            expected_output_shapes={block.outputs[0].name: (1, 5, 5, 3)},
        )

    """
    Input graph:
    input(shape=2,5)--->transpose(axis=[1,0])--->transpose(axis=[1,0])-->reduce(axis=1)
                                |                                             |
                                |                                             V
                                |                                    transpose(axis=[1,0])
                                |                                             |
                                |                                             V
                                -------------------------------------------->add------->out(shape=5,2)

    Output graph:
    input(shape=2,5)--->reduce(axis=1)---->add---->transpose(axis=[1,0])--->out(shape=5,2)
                     |                      ^
                     |                      |
                     ------------------------
    """

    def test_residual_with_unmaterialized_output(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(2, 5))])
        def prog(x):
            x1 = mb.transpose(x=x, perm=[1,0])
            t1 = mb.transpose(x=x1, perm=[1,0])
            x2 = mb.reduce_mean(x=t1, axes=[1], keep_dims=True)
            t2 = mb.transpose(x=x2, perm=[1,0])
            return mb.add(x=x1, y=t2)

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            [
                'transpose',
                'transpose',
                'reduce_mean',
                'transpose',
                'add'
            ]
        )
        self.assertEqual(
            get_op_types_in_program(prog), ['reduce_mean', 'add', 'transpose']
        )

        assert_model_is_valid(
            prog,
            {'x': (2, 5)},
            expected_output_shapes={block.outputs[0].name: (5,2)}
        )

    """
    Input graph:
    input(shape=2,5)--->transpose(axis=[1,0])--->transpose(axis=[1,0])-->reduce(axis=1)
                                |                                             |
                                |                                             V
                                |                                    transpose(axis=[1,0])
                                |                                             |
                                |                                             V
                                -------------------------------------------->add------->out1(shape=5,2)
                                                                              |
                                                                              V
                                                                            relu------->out2(shape=5,2)

    Output graph:
    input(shape=2,5)--->reduce(axis=1)----> add ----->transpose(axis=[1,0])----->out1(shape=5,2)
                     |                       |
                     |                       V
                     ---------------------> relu----->transpose(axis=[1,0])----->out2(shape=5,2)
    """

    def test_residual_with_unmaterialized_multiple_output(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(2, 5))])
        def prog(x):
            x1 = mb.transpose(x=x, perm=[1,0])
            t1 = mb.transpose(x=x1, perm=[1,0])
            x2 = mb.reduce_mean(x=t1, axes=[1], keep_dims=True)
            t2 = mb.transpose(x=x2, perm=[1,0])
            out1 = mb.add(x=x1, y=t2)
            out2 = mb.relu(x=out1)
            return out1, out2

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            [
                'transpose',
                'transpose',
                'reduce_mean',
                'transpose',
                'add',
                'relu'
            ]
        )
        self.assertEqual(
            get_op_types_in_program(prog), ['reduce_mean', 'add', 'relu', 'transpose', 'transpose']
        )

        assert_model_is_valid(
            prog,
            {'x': (2, 5)},
            expected_output_shapes={block.outputs[0].name: (5,2),
                                    block.outputs[1].name: (5,2)}
        )

    """
    Input graph:
    input(shape=2,5)---->transpose(axis=[1,0])------>relu----->transpose(axis=[1,0])------>out2(shape=2,5)
                                                       |
                                                       ------->out1(shape=5,2)

    Output graph:
    input(shape=2,5)---->relu-----> out2(shape=2,5)
                           |
                           V
                    transpose(axis=[1,0]) -----> out1(shape=5,2)
    """

    def test_materialized_output_reuse(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(2, 5))])
        def prog(x):
            x1 = mb.transpose(x=x, perm=[1,0])
            y1 = mb.relu(x=x1)
            y2 = mb.transpose(x=y1, perm=[1,0])
            return y1, y2

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            [
                'transpose',
                'relu',
                'transpose',
            ]
        )
        self.assertEqual(
            get_op_types_in_program(prog), ['relu', 'transpose']
        )

        assert_model_is_valid(
            prog,
            {'x': (2, 5)},
            expected_output_shapes={block.outputs[0].name: (5,2),
                                    block.outputs[1].name: (2,5)}
        )

    """
    Input graph:
    input(shape=1,2,5,5)----->transpose(axis=[0,2,3,1])------->add------------>transpose(axis=[0,3,1,2])--->out1(shape=1,2,5,5)
                                                          |     ^        |
                                                          |     |        |
                                                          ---->relu      ----->transpose(axis=[0,3,1,2])--->out2(shape=1,2,5,5)

    Output graph:
    input(shape=1,2,5,5)----->add------->out1(shape=1,2,5,5)
                      |        ^    |
                      |        |    |
                      |------>relu  ------identity(renaming)---->out2(shape=1,2,5,5)
    """

    def test_fusion_with_double_outputs(self):
        @mb.program(input_specs=[mb.TensorSpec(shape=(1, 2, 5, 5))])
        def prog(x):
            x1 = mb.transpose(x=x, perm=[0, 2, 3, 1])
            x2 = mb.relu(x=x1)
            x3 = mb.add(x=x1, y=x2)
            y1 = mb.transpose(x=x3, perm=[0, 3, 1, 2])
            y2 = mb.transpose(x=x3, perm=[0, 3, 1, 2])
            return y1, y2

        prev_prog, prev_block, block = apply_pass_and_basic_check(
            prog, "common::reduce_transposes"
        )
        self.assertEqual(
            get_op_types_in_program(prev_prog),
            ["transpose", "relu", "add", "transpose", "transpose"],
        )
        self.assertEqual(get_op_types_in_program(prog), ["relu", "add", "identity"])

        assert block.find_ops(op_type="relu")[0].inputs["x"] == block.inputs["x"]
        assert block.find_ops(op_type="add")[0].inputs["x"] == block.inputs["x"]
        assert (
            block.find_ops(op_type="add")[0].inputs["y"]
            == block.find_ops(op_type="relu")[0].outputs[0]
        )

        assert_model_is_valid(
            prog,
            {"x": (1, 2, 5, 5)},
            expected_output_shapes={block.outputs[0].name: (1, 2, 5, 5)},
        )




#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil import testing_reqs
from coremltools.converters.mil.testing_reqs import *

from .testing_utils import run_compare_builder

backends = testing_reqs.backends


class TestElementwiseBinary:
    # All in this test share the same backends
    @pytest.mark.parametrize(
        "use_cpu_only, backend, mode",
        itertools.product(
            [True, False],
            backends,
            [
                "add",
                "floor_div",
                "maximum",
                "minimum",
                "mod",
                "mul",
                "pow",
                "real_div",
                "sub",
            ],
        ),
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend, mode):
        if mode == "add":
            x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
            y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array([[0, 4, 0], [8, 0, 12]], dtype=np.float32)

            build = lambda x, y: mb.add(x=x, y=y)
        elif mode == "floor_div":
            x = np.array([[10, 20, 30], [40, 50, 60]], dtype=np.float32)
            y = np.array([[11, 12, 13], [14, 15, 16]], dtype=np.float32)
            expected_outputs = np.array([[0, 1, 2], [2, 3, 3]], dtype=np.float32)

            build = lambda x, y: mb.floor_div(x=x, y=y)
        elif mode == "maximum":
            x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
            y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)

            build = lambda x, y: mb.maximum(x=x, y=y)
        elif mode == "minimum":
            x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
            y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)

            build = lambda x, y: mb.minimum(x=x, y=y)
        elif mode == "mod":
            x = np.array([[10, 20, 30], [40, 50, 60]], dtype=np.float32)
            y = np.array([[11, 12, 13], [14, 15, 16]], dtype=np.float32)
            expected_outputs = np.array([[10, 8, 4], [12, 5, 12]], dtype=np.float32)

            build = lambda x, y: mb.mod(x=x, y=y)
        elif mode == "mul":
            x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
            y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array([[-1, 4, -9], [16, -25, 36]], dtype=np.float32)

            build = lambda x, y: mb.mul(x=x, y=y)
        elif mode == "pow":
            x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
            y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array(
                [[1, 4, 0.037], [256, 0.00032, 46656]], dtype=np.float32
            )

            build = lambda x, y: mb.pow(x=x, y=y)
        elif mode == "real_div":
            x = np.array([[10, 20, 30], [40, 50, 60]], dtype=np.float32)
            y = np.array([[11, 12, 13], [14, 15, 16]], dtype=np.float32)
            expected_outputs = np.array(
                [[0.90909091, 1.66666667, 2.30769231], [2.85714286, 3.33333333, 3.75]],
                dtype=np.float32,
            )

            build = lambda x, y: mb.real_div(x=x, y=y)
        elif mode == "sub":
            x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
            y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            expected_outputs = np.array([[2, 0, 6], [0, 10, 0]], dtype=np.float32)

            build = lambda x, y: mb.sub(x=x, y=y)

        expected_output_types = (2, 3, types.fp32)
        input_placeholders = {
            "x": mb.placeholder(shape=x.shape),
            "y": mb.placeholder(shape=y.shape),
        }
        input_values = {"x": x, "y": y}
        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_add(self):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        expected_outputs = np.array([[0, 4, 0], [8, 0, 12]], dtype=np.float32)
        v = mb.add(x=x, y=y)
        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_floor_div(self):
        x = np.array([[10, 20, 30], [40, 50, 60]], dtype=np.float32)
        y = np.array([[11, 12, 13], [14, 15, 16]], dtype=np.float32)
        expected_outputs = np.array([[0, 1, 2], [2, 3, 3]], dtype=np.float32)
        v = mb.floor_div(x=x, y=y)
        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_maximum(self):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        expected_outputs = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        v = mb.maximum(x=x, y=y)
        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_minimum(self):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        expected_outputs = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        v = mb.minimum(x=x, y=y)
        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_mod(self):
        x = np.array([[10, 20, 30], [40, 50, 60]], dtype=np.float32)
        y = np.array([[11, 12, 13], [14, 15, 16]], dtype=np.float32)
        expected_outputs = np.array([[10, 8, 4], [12, 5, 12]], dtype=np.float32)
        v = mb.mod(x=x, y=y)
        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_mul(self):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        expected_outputs = np.array([[-1, 4, -9], [16, -25, 36]], dtype=np.float32)
        v = mb.mul(x=x, y=y)
        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_pow(self):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        expected_outputs = np.array(
            [[1, 4, 0.037], [256, 0.00032, 46656]], dtype=np.float32
        )
        v = mb.pow(x=x, y=y)
        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_real_div(self):
        x = np.array([[10, 20, 30], [40, 50, 60]], dtype=np.float32)
        y = np.array([[11, 12, 13], [14, 15, 16]], dtype=np.float32)
        expected_outputs = np.array(
            [[0.90909091, 1.66666667, 2.30769231], [2.85714286, 3.33333333, 3.75]],
            dtype=np.float32,
        )
        v = mb.real_div(x=x, y=y)
        assert is_close(expected_outputs, v.val)

    @ssa_fn
    def test_builder_real_div_both_ints(self):
        x = np.array([5], dtype=np.int32)
        y = np.array([2], dtype=np.int32)
        expected_outputs = np.array([2.5], dtype=np.float32)
        v = mb.real_div(x=x, y=y)
        assert is_close(expected_outputs, v.val)
        # real_div should produce float values regardless of input type
        assert isinstance(v.val[0], (float, np.float32))
        # make sure the dtype is float
        assert types.is_float(v.dtype)
        # make sure the symbolic type matches the value type
        assert v._sym_type.get_primitive() == v._sym_val.get_primitive()

    @ssa_fn
    def test_builder_sub(self):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        expected_outputs = np.array([[2, 0, 6], [0, 10, 0]], dtype=np.float32)
        v = mb.sub(x=x, y=y)
        assert is_close(expected_outputs, v.val)


class TestEqual:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {
            "x": mb.placeholder(shape=x.shape),
            "y": mb.placeholder(shape=y.shape),
        }
        input_values = {"x": x, "y": y}

        def build(x, y):
            return mb.equal(x=x, y=y), mb.equal(x=-3, y=y)

        expected_output_types = [
            (2, 3, types.bool),
            (2, 3, types.bool),
        ]
        expected_outputs = [
            np.array([[0, 1, 0], [1, 0, 1]], dtype=np.bool),
            np.array([[0, 0, 1], [0, 0, 0]], dtype=np.bool),
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        expected_outputs = np.array([[0, 1, 0], [1, 0, 1]], dtype=np.bool)
        v = mb.equal(x=x_val, y=y_val)
        assert is_close(expected_outputs, v.val)


class TestGreater:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {
            "x": mb.placeholder(shape=x.shape),
            "y": mb.placeholder(shape=y.shape),
        }
        input_values = {"x": x, "y": y}

        def build(x, y):
            return mb.greater(x=x, y=y), mb.greater(x=x, y=3.5)

        expected_output_types = [
            (2, 3, types.bool),
            (2, 3, types.bool),
        ]
        expected_outputs = [
            np.array([[1, 0, 1], [0, 1, 0]], dtype=np.bool),
            np.array([[0, 0, 0], [1, 1, 1]], dtype=np.bool),
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        expected_outputs = np.array([[1, 0, 1], [0, 1, 0]], dtype=np.bool)
        v = mb.greater(x=x_val, y=y_val)
        assert is_close(expected_outputs, v.val)


class TestGreaterEqual:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {
            "x": mb.placeholder(shape=x.shape),
            "y": mb.placeholder(shape=y.shape),
        }
        input_values = {"x": x, "y": y}

        def build(x, y):
            return mb.greater_equal(x=x, y=y), mb.greater_equal(x=x, y=3.5)

        expected_output_types = [
            (2, 3, types.bool),
            (2, 3, types.bool),
        ]
        expected_outputs = [
            np.array([[1, 1, 1], [1, 1, 1]], dtype=np.bool),
            np.array([[0, 0, 0], [1, 1, 1]], dtype=np.bool),
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        expected_outputs = np.array([[1, 1, 1], [1, 1, 1]], dtype=np.bool)
        v = mb.greater_equal(x=x_val, y=y_val)
        assert is_close(expected_outputs, v.val)


class TestLess:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {
            "x": mb.placeholder(shape=x.shape),
            "y": mb.placeholder(shape=y.shape),
        }
        input_values = {"x": x, "y": y}

        def build(x, y):
            return mb.less(x=x, y=y)

        expected_output_types = (2, 3, types.bool)
        expected_outputs = np.array([[0, 0, 0], [0, 0, 0]], dtype=np.bool)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke2(self, use_cpu_only, backend):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=x.shape)}
        input_values = {"x": x}

        def build(x):
            # y is const
            y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
            return mb.less(x=x, y=y)

        expected_output_types = (2, 3, types.bool)
        expected_outputs = np.array([[0, 0, 0], [0, 0, 0]], dtype=np.bool)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_broadcast(self, use_cpu_only, backend):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=x.shape)}
        input_values = {"x": x}

        def build(x):
            # y is const
            return mb.less(x=x, y=3.5)

        expected_output_types = (2, 3, types.bool)
        expected_outputs = np.array([[1, 1, 1], [0, 0, 0]], dtype=np.bool)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        expected_outputs = np.array([[0, 0, 0], [0, 0, 0]], dtype=np.bool)
        v = mb.less(x=x_val, y=y_val)
        assert is_close(expected_outputs, v.val)


class TestLessEqual:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {
            "x": mb.placeholder(shape=x.shape),
            "y": mb.placeholder(shape=y.shape),
        }
        input_values = {"x": x, "y": y}

        def build(x, y):
            return mb.less_equal(x=x, y=y)

        expected_output_types = (2, 3, types.bool)
        expected_outputs = np.array([[0, 1, 0], [1, 0, 1]], dtype=np.bool)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        expected_outputs = np.array([[0, 1, 0], [1, 0, 1]], dtype=np.bool)
        v = mb.less_equal(x=x_val, y=y_val)
        assert is_close(expected_outputs, v.val)


class TestNotEqual:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        input_placeholders = {
            "x": mb.placeholder(shape=x.shape),
            "y": mb.placeholder(shape=y.shape),
        }
        input_values = {"x": x, "y": y}

        def build(x, y):
            return mb.not_equal(x=x, y=y)

        expected_output_types = (2, 3, types.bool)
        expected_outputs = np.array([[1, 0, 1], [0, 1, 0]], dtype=np.bool)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        y_val = np.array([[-1, 2, -3], [4, -5, 6]], dtype=np.float32)
        expected_outputs = np.array([[1, 0, 1], [0, 1, 0]], dtype=np.bool)
        v = mb.not_equal(x=x_val, y=y_val)
        assert is_close(expected_outputs, v.val)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: DictVectorizer.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import DataStructures_pb2 as DataStructures__pb2
try:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes_pb2

from .DataStructures_pb2 import *

DESCRIPTOR = _descriptor.FileDescriptor(
  name='DictVectorizer.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x14\x44ictVectorizer.proto\x12\x14\x43oreML.Specification\x1a\x14\x44\x61taStructures.proto\"\x8f\x01\n\x0e\x44ictVectorizer\x12;\n\rstringToIndex\x18\x01 \x01(\x0b\x32\".CoreML.Specification.StringVectorH\x00\x12\x39\n\x0cint64ToIndex\x18\x02 \x01(\x0b\x32!.CoreML.Specification.Int64VectorH\x00\x42\x05\n\x03MapB\x02H\x03P\x00\x62\x06proto3')
  ,
  dependencies=[DataStructures__pb2.DESCRIPTOR,],
  public_dependencies=[DataStructures__pb2.DESCRIPTOR,])




_DICTVECTORIZER = _descriptor.Descriptor(
  name='DictVectorizer',
  full_name='CoreML.Specification.DictVectorizer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='stringToIndex', full_name='CoreML.Specification.DictVectorizer.stringToIndex', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='int64ToIndex', full_name='CoreML.Specification.DictVectorizer.int64ToIndex', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Map', full_name='CoreML.Specification.DictVectorizer.Map',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=69,
  serialized_end=212,
)

_DICTVECTORIZER.fields_by_name['stringToIndex'].message_type = DataStructures__pb2._STRINGVECTOR
_DICTVECTORIZER.fields_by_name['int64ToIndex'].message_type = DataStructures__pb2._INT64VECTOR
_DICTVECTORIZER.oneofs_by_name['Map'].fields.append(
  _DICTVECTORIZER.fields_by_name['stringToIndex'])
_DICTVECTORIZER.fields_by_name['stringToIndex'].containing_oneof = _DICTVECTORIZER.oneofs_by_name['Map']
_DICTVECTORIZER.oneofs_by_name['Map'].fields.append(
  _DICTVECTORIZER.fields_by_name['int64ToIndex'])
_DICTVECTORIZER.fields_by_name['int64ToIndex'].containing_oneof = _DICTVECTORIZER.oneofs_by_name['Map']
DESCRIPTOR.message_types_by_name['DictVectorizer'] = _DICTVECTORIZER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DictVectorizer = _reflection.GeneratedProtocolMessageType('DictVectorizer', (_message.Message,), dict(
  DESCRIPTOR = _DICTVECTORIZER,
  __module__ = 'DictVectorizer_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.DictVectorizer)
  ))
_sym_db.RegisterMessage(DictVectorizer)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)
